"""
Enterprise Dependency Injection - IoC container for agent components.

Provides automatic wiring of dependencies, scoped lifetimes, and 
easy testability through dependency injection.

Features:
- Constructor injection
- Decorator-based registration
- Scopes (singleton, transient, scoped)
- Factory functions
- Auto-wiring
- Named dependencies
"""

import asyncio
import functools
import inspect
import threading
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass, field
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

T = TypeVar("T")


# =============================================================================
# Lifetimes
# =============================================================================

class Lifetime(Enum):
    """Dependency lifetime scopes."""
    TRANSIENT = "transient"      # New instance every time
    SINGLETON = "singleton"       # Single instance for container
    SCOPED = "scoped"            # One instance per scope


# =============================================================================
# Registration
# =============================================================================

@dataclass
class Registration:
    """A dependency registration."""
    service_type: Type
    implementation: Union[Type, Callable]
    lifetime: Lifetime
    name: Optional[str] = None
    factory: Optional[Callable] = None
    instance: Optional[Any] = None


# =============================================================================
# Container
# =============================================================================

class Container:
    """
    Dependency injection container.
    
    Usage:
        >>> container = Container()
        >>> 
        >>> @container.register(Lifetime.SINGLETON)
        >>> class ConfigService:
        ...     def __init__(self):
        ...         self.debug = True
        >>> 
        >>> @container.register(Lifetime.TRANSIENT)
        >>> class AgentService:
        ...     def __init__(self, config: ConfigService):
        ...         self.config = config
        >>> 
        >>> agent = container.resolve(AgentService)
    """
    
    def __init__(self, parent: Optional["Container"] = None):
        self._registrations: Dict[str, Registration] = {}
        self._singletons: Dict[str, Any] = {}
        self._parent = parent
        self._lock = threading.RLock()
    
    def _get_key(self, service_type: Type, name: Optional[str] = None) -> str:
        """Get registration key."""
        type_name = f"{service_type.__module__}.{service_type.__qualname__}"
        if name:
            return f"{type_name}:{name}"
        return type_name
    
    def register(
        self,
        lifetime: Lifetime = Lifetime.TRANSIENT,
        name: Optional[str] = None,
        interface: Optional[Type] = None,
    ):
        """
        Decorator to register a class.
        
        Args:
            lifetime: Dependency lifetime
            name: Optional name for multiple implementations
            interface: Optional interface type to register as
        """
        def decorator(cls: Type[T]) -> Type[T]:
            service_type = interface or cls
            key = self._get_key(service_type, name)
            
            with self._lock:
                self._registrations[key] = Registration(
                    service_type=service_type,
                    implementation=cls,
                    lifetime=lifetime,
                    name=name,
                )
            
            return cls
        return decorator
    
    def register_instance(
        self,
        service_type: Type[T],
        instance: T,
        name: Optional[str] = None,
    ):
        """Register an existing instance as singleton."""
        key = self._get_key(service_type, name)
        
        with self._lock:
            self._registrations[key] = Registration(
                service_type=service_type,
                implementation=type(instance),
                lifetime=Lifetime.SINGLETON,
                name=name,
                instance=instance,
            )
            self._singletons[key] = instance
    
    def register_factory(
        self,
        service_type: Type[T],
        factory: Callable[..., T],
        lifetime: Lifetime = Lifetime.TRANSIENT,
        name: Optional[str] = None,
    ):
        """Register a factory function."""
        key = self._get_key(service_type, name)
        
        with self._lock:
            self._registrations[key] = Registration(
                service_type=service_type,
                implementation=factory,
                lifetime=lifetime,
                name=name,
                factory=factory,
            )
    
    def register_type(
        self,
        service_type: Type,
        implementation: Type,
        lifetime: Lifetime = Lifetime.TRANSIENT,
        name: Optional[str] = None,
    ):
        """Register an implementation for an interface."""
        key = self._get_key(service_type, name)
        
        with self._lock:
            self._registrations[key] = Registration(
                service_type=service_type,
                implementation=implementation,
                lifetime=lifetime,
                name=name,
            )
    
    def resolve(
        self,
        service_type: Type[T],
        name: Optional[str] = None,
    ) -> T:
        """
        Resolve a dependency.
        
        Args:
            service_type: Type to resolve
            name: Optional name for named dependencies
            
        Returns:
            Instance of the requested type
        """
        key = self._get_key(service_type, name)
        
        # Check for existing singleton
        if key in self._singletons:
            return self._singletons[key]
        
        # Check parent
        if self._parent and key not in self._registrations:
            return self._parent.resolve(service_type, name)
        
        # Get registration
        registration = self._registrations.get(key)
        
        if not registration:
            # Try auto-registration
            if inspect.isclass(service_type) and not inspect.isabstract(service_type):
                self.register_type(service_type, service_type, Lifetime.TRANSIENT)
                registration = self._registrations[key]
            else:
                raise DependencyError(f"No registration for {service_type.__name__}")
        
        return self._create_instance(registration)
    
    def _create_instance(self, registration: Registration) -> Any:
        """Create an instance from a registration."""
        key = self._get_key(registration.service_type, registration.name)
        
        # Return existing singleton
        if registration.lifetime == Lifetime.SINGLETON:
            if registration.instance:
                self._singletons[key] = registration.instance
                return registration.instance
            
            if key in self._singletons:
                return self._singletons[key]
        
        # Use factory if provided
        if registration.factory:
            deps = self._resolve_dependencies(registration.factory)
            instance = registration.factory(**deps)
        else:
            # Create via constructor
            impl = registration.implementation
            deps = self._resolve_dependencies(impl)
            instance = impl(**deps)
        
        # Store singleton
        if registration.lifetime == Lifetime.SINGLETON:
            with self._lock:
                self._singletons[key] = instance
        
        return instance
    
    def _resolve_dependencies(
        self,
        target: Union[Type, Callable],
    ) -> Dict[str, Any]:
        """Resolve constructor/function dependencies."""
        deps = {}
        
        try:
            hints = get_type_hints(target)
        except Exception:
            hints = {}
        
        sig = inspect.signature(target)
        
        for param_name, param in sig.parameters.items():
            if param_name == "self":
                continue
            
            if param_name in hints:
                param_type = hints[param_name]
                
                # Handle Optional types
                origin = get_origin(param_type)
                if origin is Union:
                    args = get_args(param_type)
                    non_none = [a for a in args if a is not type(None)]
                    if non_none:
                        param_type = non_none[0]
                
                # Try to resolve
                try:
                    deps[param_name] = self.resolve(param_type)
                except DependencyError:
                    if param.default is param.empty:
                        raise
        
        return deps
    
    def create_scope(self) -> "Scope":
        """Create a new dependency scope."""
        return Scope(self)
    
    def get_all(self, service_type: Type[T]) -> List[T]:
        """Get all implementations of a type."""
        instances = []
        type_name = f"{service_type.__module__}.{service_type.__qualname__}"
        
        for key, reg in self._registrations.items():
            if key.startswith(type_name):
                instances.append(self._create_instance(reg))
        
        return instances
    
    def is_registered(
        self,
        service_type: Type,
        name: Optional[str] = None,
    ) -> bool:
        """Check if a type is registered."""
        key = self._get_key(service_type, name)
        return key in self._registrations


class Scope:
    """
    A dependency scope for scoped lifetimes.
    
    Usage:
        >>> with container.create_scope() as scope:
        ...     db = scope.resolve(DatabaseConnection)
        ...     # Same instance within scope
    """
    
    def __init__(self, container: Container):
        self._container = container
        self._scoped: Dict[str, Any] = {}
    
    def __enter__(self) -> "Scope":
        return self
    
    def __exit__(self, *args):
        self._dispose()
    
    async def __aenter__(self) -> "Scope":
        return self
    
    async def __aexit__(self, *args):
        self._dispose()
    
    def resolve(
        self,
        service_type: Type[T],
        name: Optional[str] = None,
    ) -> T:
        """Resolve a dependency within this scope."""
        key = self._container._get_key(service_type, name)
        
        # Check scoped cache
        if key in self._scoped:
            return self._scoped[key]
        
        # Get registration
        registration = self._container._registrations.get(key)
        
        if registration and registration.lifetime == Lifetime.SCOPED:
            instance = self._container._create_instance(registration)
            self._scoped[key] = instance
            return instance
        
        return self._container.resolve(service_type, name)
    
    def _dispose(self):
        """Dispose scoped instances."""
        for instance in self._scoped.values():
            if hasattr(instance, "dispose"):
                instance.dispose()
            elif hasattr(instance, "close"):
                instance.close()
        self._scoped.clear()


# =============================================================================
# Decorators
# =============================================================================

def inject(fn: Callable) -> Callable:
    """
    Decorator to inject dependencies into a function.
    
    Usage:
        >>> @inject
        >>> def process(config: ConfigService, data: str):
        ...     return config.process(data)
    """
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        container = get_container()
        deps = container._resolve_dependencies(fn)
        
        # Don't override explicit args
        for key in list(deps.keys()):
            if key in kwargs:
                del deps[key]
        
        return fn(*args, **deps, **kwargs)
    
    return wrapper


def inject_async(fn: Callable) -> Callable:
    """Async version of inject decorator."""
    @functools.wraps(fn)
    async def wrapper(*args, **kwargs):
        container = get_container()
        deps = container._resolve_dependencies(fn)
        
        for key in list(deps.keys()):
            if key in kwargs:
                del deps[key]
        
        return await fn(*args, **deps, **kwargs)
    
    return wrapper


# =============================================================================
# Service Provider Interface
# =============================================================================

class ServiceProvider(ABC):
    """Base class for service providers."""
    
    @abstractmethod
    def register(self, container: Container):
        """Register services in the container."""
        pass


class ServiceCollection:
    """
    Builder for configuring services.
    
    Usage:
        >>> services = ServiceCollection()
        >>> services.add_singleton(ConfigService)
        >>> services.add_transient(AgentService)
        >>> container = services.build()
    """
    
    def __init__(self):
        self._registrations: List[tuple] = []
    
    def add_singleton(
        self,
        service_type: Type[T],
        implementation: Optional[Type[T]] = None,
        name: Optional[str] = None,
    ) -> "ServiceCollection":
        """Add a singleton service."""
        self._registrations.append((
            service_type,
            implementation or service_type,
            Lifetime.SINGLETON,
            name,
        ))
        return self
    
    def add_transient(
        self,
        service_type: Type[T],
        implementation: Optional[Type[T]] = None,
        name: Optional[str] = None,
    ) -> "ServiceCollection":
        """Add a transient service."""
        self._registrations.append((
            service_type,
            implementation or service_type,
            Lifetime.TRANSIENT,
            name,
        ))
        return self
    
    def add_scoped(
        self,
        service_type: Type[T],
        implementation: Optional[Type[T]] = None,
        name: Optional[str] = None,
    ) -> "ServiceCollection":
        """Add a scoped service."""
        self._registrations.append((
            service_type,
            implementation or service_type,
            Lifetime.SCOPED,
            name,
        ))
        return self
    
    def add_factory(
        self,
        service_type: Type[T],
        factory: Callable[..., T],
        lifetime: Lifetime = Lifetime.TRANSIENT,
        name: Optional[str] = None,
    ) -> "ServiceCollection":
        """Add a factory registration."""
        self._registrations.append((
            service_type,
            factory,
            lifetime,
            name,
            True,  # is_factory
        ))
        return self
    
    def add_instance(
        self,
        service_type: Type[T],
        instance: T,
        name: Optional[str] = None,
    ) -> "ServiceCollection":
        """Add an existing instance."""
        self._registrations.append((
            service_type,
            instance,
            Lifetime.SINGLETON,
            name,
            False,  # is_factory
            True,   # is_instance
        ))
        return self
    
    def add_provider(self, provider: ServiceProvider) -> "ServiceCollection":
        """Add a service provider."""
        self._registrations.append(("provider", provider))
        return self
    
    def build(self) -> Container:
        """Build the container from registrations."""
        container = Container()
        
        for reg in self._registrations:
            if reg[0] == "provider":
                reg[1].register(container)
            elif len(reg) > 5 and reg[5]:  # is_instance
                container.register_instance(reg[0], reg[1], reg[3])
            elif len(reg) > 4 and reg[4]:  # is_factory
                container.register_factory(reg[0], reg[1], reg[2], reg[3])
            else:
                container.register_type(reg[0], reg[1], reg[2], reg[3])
        
        return container


# =============================================================================
# Errors
# =============================================================================

class DependencyError(Exception):
    """Error resolving dependency."""
    pass


# =============================================================================
# Global Container
# =============================================================================

_global_container: Optional[Container] = None
_lock = threading.Lock()


def get_container() -> Container:
    """Get the global container."""
    global _global_container
    
    if _global_container is None:
        with _lock:
            if _global_container is None:
                _global_container = Container()
    
    return _global_container


def set_container(container: Container):
    """Set the global container."""
    global _global_container
    _global_container = container


def reset_container():
    """Reset the global container."""
    global _global_container
    _global_container = None


# Convenience functions
def singleton(name: Optional[str] = None, interface: Optional[Type] = None):
    """Register as singleton in global container."""
    return get_container().register(Lifetime.SINGLETON, name, interface)


def transient(name: Optional[str] = None, interface: Optional[Type] = None):
    """Register as transient in global container."""
    return get_container().register(Lifetime.TRANSIENT, name, interface)


def scoped(name: Optional[str] = None, interface: Optional[Type] = None):
    """Register as scoped in global container."""
    return get_container().register(Lifetime.SCOPED, name, interface)


def resolve(service_type: Type[T], name: Optional[str] = None) -> T:
    """Resolve from global container."""
    return get_container().resolve(service_type, name)


# =============================================================================
# Pre-built Service Providers
# =============================================================================

class CoreServicesProvider(ServiceProvider):
    """Core framework services."""
    
    def register(self, container: Container):
        from ..config import ConfigManager
        from ..monitoring import MonitoringManager
        
        container.register_type(ConfigManager, ConfigManager, Lifetime.SINGLETON)
        container.register_type(MonitoringManager, MonitoringManager, Lifetime.SINGLETON)


class LLMServicesProvider(ServiceProvider):
    """LLM-related services."""
    
    def __init__(self, model: str = "gpt-4o"):
        self.model = model
    
    def register(self, container: Container):
        # Register LLM factory
        container.register_factory(
            type(None),  # Will be replaced with actual LLM interface
            lambda: self._create_llm(),
            Lifetime.SINGLETON,
            "llm",
        )
    
    def _create_llm(self):
        from ..llms import create_llm
        return create_llm(self.model)


# =============================================================================
# Context Manager Integration
# =============================================================================

@contextmanager
def scope():
    """Context manager for dependency scope."""
    container = get_container()
    with container.create_scope() as s:
        yield s


@asynccontextmanager
async def async_scope():
    """Async context manager for dependency scope."""
    container = get_container()
    async with container.create_scope() as s:
        yield s



"""
Enterprise SLA Manager Module.

SLA tracking, uptime monitoring, breach alerts,
and service level reporting.

Example:
    # Create SLA manager
    sla = create_sla_manager()
    
    # Define SLA
    await sla.define(
        name="api_availability",
        target=99.9,
        metric_type=MetricType.UPTIME,
        period=SLAPeriod.MONTHLY,
    )
    
    # Record metric
    await sla.record(
        sla_name="api_availability",
        value=100.0,
        timestamp=datetime.utcnow(),
    )
    
    # Check compliance
    status = await sla.get_status("api_availability")
"""

from __future__ import annotations

import asyncio
import logging
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from decimal import Decimal
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

logger = logging.getLogger(__name__)


class SLAError(Exception):
    """SLA error."""
    pass


class SLABreach(SLAError):
    """SLA breach."""
    pass


class MetricType(str, Enum):
    """Metric type."""
    UPTIME = "uptime"
    LATENCY = "latency"
    ERROR_RATE = "error_rate"
    THROUGHPUT = "throughput"
    RESPONSE_TIME = "response_time"
    AVAILABILITY = "availability"
    MTTR = "mttr"  # Mean time to recover
    MTBF = "mtbf"  # Mean time between failures
    CUSTOM = "custom"


class SLAPeriod(str, Enum):
    """SLA period."""
    HOURLY = "hourly"
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    QUARTERLY = "quarterly"
    YEARLY = "yearly"


class SLAStatus(str, Enum):
    """SLA status."""
    COMPLIANT = "compliant"
    AT_RISK = "at_risk"
    BREACHED = "breached"
    UNKNOWN = "unknown"


class AlertSeverity(str, Enum):
    """Alert severity."""
    INFO = "info"
    WARNING = "warning"
    CRITICAL = "critical"


@dataclass
class MetricRecord:
    """Metric record."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    sla_id: str = ""
    value: float = 0.0
    timestamp: datetime = field(default_factory=datetime.utcnow)
    source: str = ""
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class SLADefinition:
    """SLA definition."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    description: str = ""
    metric_type: MetricType = MetricType.UPTIME
    target: float = 99.9
    warning_threshold: float = 99.5
    period: SLAPeriod = SLAPeriod.MONTHLY
    
    # Comparison
    higher_is_better: bool = True  # For latency, lower is better
    
    # Service info
    service: str = ""
    owner: str = ""
    
    # Enabled
    enabled: bool = True
    
    # Dates
    created_at: datetime = field(default_factory=datetime.utcnow)
    effective_from: Optional[datetime] = None
    effective_until: Optional[datetime] = None
    
    # Metadata
    tags: Dict[str, str] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class SLAStatusReport:
    """SLA status report."""
    sla_id: str = ""
    sla_name: str = ""
    status: SLAStatus = SLAStatus.UNKNOWN
    current_value: float = 0.0
    target: float = 0.0
    variance: float = 0.0
    period_start: datetime = field(default_factory=datetime.utcnow)
    period_end: datetime = field(default_factory=datetime.utcnow)
    records_count: int = 0
    is_breached: bool = False
    breach_duration: Optional[timedelta] = None
    error_budget_remaining: float = 100.0


@dataclass
class SLAAlert:
    """SLA alert."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    sla_id: str = ""
    sla_name: str = ""
    severity: AlertSeverity = AlertSeverity.WARNING
    status: SLAStatus = SLAStatus.AT_RISK
    current_value: float = 0.0
    target: float = 0.0
    message: str = ""
    triggered_at: datetime = field(default_factory=datetime.utcnow)
    acknowledged: bool = False
    resolved: bool = False


@dataclass
class ErrorBudget:
    """Error budget."""
    sla_id: str = ""
    total_minutes: float = 0.0
    consumed_minutes: float = 0.0
    remaining_minutes: float = 0.0
    remaining_percentage: float = 100.0
    period_start: datetime = field(default_factory=datetime.utcnow)
    period_end: datetime = field(default_factory=datetime.utcnow)


@dataclass
class SLAStats:
    """SLA statistics."""
    total_slas: int = 0
    compliant: int = 0
    at_risk: int = 0
    breached: int = 0
    avg_compliance: float = 0.0


# SLA store
class SLAStore(ABC):
    """SLA storage."""
    
    @abstractmethod
    async def save(self, sla: SLADefinition) -> None:
        pass
    
    @abstractmethod
    async def get(self, sla_id: str) -> Optional[SLADefinition]:
        pass
    
    @abstractmethod
    async def get_by_name(self, name: str) -> Optional[SLADefinition]:
        pass
    
    @abstractmethod
    async def list_all(self) -> List[SLADefinition]:
        pass
    
    @abstractmethod
    async def delete(self, sla_id: str) -> bool:
        pass


class InMemorySLAStore(SLAStore):
    """In-memory SLA store."""
    
    def __init__(self):
        self._slas: Dict[str, SLADefinition] = {}
        self._by_name: Dict[str, str] = {}
    
    async def save(self, sla: SLADefinition) -> None:
        self._slas[sla.id] = sla
        self._by_name[sla.name] = sla.id
    
    async def get(self, sla_id: str) -> Optional[SLADefinition]:
        return self._slas.get(sla_id)
    
    async def get_by_name(self, name: str) -> Optional[SLADefinition]:
        sla_id = self._by_name.get(name)
        if sla_id:
            return self._slas.get(sla_id)
        return None
    
    async def list_all(self) -> List[SLADefinition]:
        return list(self._slas.values())
    
    async def delete(self, sla_id: str) -> bool:
        sla = self._slas.get(sla_id)
        if sla:
            del self._slas[sla_id]
            self._by_name.pop(sla.name, None)
            return True
        return False


# Metric store
class MetricStore(ABC):
    """Metric storage."""
    
    @abstractmethod
    async def save(self, record: MetricRecord) -> None:
        pass
    
    @abstractmethod
    async def list_by_sla(
        self,
        sla_id: str,
        start_date: datetime,
        end_date: datetime,
    ) -> List[MetricRecord]:
        pass
    
    @abstractmethod
    async def get_average(
        self,
        sla_id: str,
        start_date: datetime,
        end_date: datetime,
    ) -> Optional[float]:
        pass


class InMemoryMetricStore(MetricStore):
    """In-memory metric store."""
    
    def __init__(self, max_records: int = 1000000):
        self._records: Dict[str, MetricRecord] = {}
        self._by_sla: Dict[str, List[str]] = {}
        self._max_records = max_records
    
    async def save(self, record: MetricRecord) -> None:
        self._records[record.id] = record
        
        if record.sla_id not in self._by_sla:
            self._by_sla[record.sla_id] = []
        
        self._by_sla[record.sla_id].append(record.id)
        
        # Trim old records
        if len(self._records) > self._max_records:
            oldest = list(self._records.keys())[0]
            del self._records[oldest]
    
    async def list_by_sla(
        self,
        sla_id: str,
        start_date: datetime,
        end_date: datetime,
    ) -> List[MetricRecord]:
        record_ids = self._by_sla.get(sla_id, [])
        records = []
        
        for record_id in record_ids:
            record = self._records.get(record_id)
            if record:
                if start_date <= record.timestamp <= end_date:
                    records.append(record)
        
        return sorted(records, key=lambda r: r.timestamp)
    
    async def get_average(
        self,
        sla_id: str,
        start_date: datetime,
        end_date: datetime,
    ) -> Optional[float]:
        records = await self.list_by_sla(sla_id, start_date, end_date)
        
        if not records:
            return None
        
        return sum(r.value for r in records) / len(records)


# Period calculator
class PeriodCalculator:
    """SLA period calculator."""
    
    @staticmethod
    def get_period_range(
        period: SLAPeriod,
        reference: Optional[datetime] = None,
    ) -> Tuple[datetime, datetime]:
        """Get period start and end dates."""
        ref = reference or datetime.utcnow()
        
        if period == SLAPeriod.HOURLY:
            start = ref.replace(minute=0, second=0, microsecond=0)
            end = start + timedelta(hours=1) - timedelta(microseconds=1)
        
        elif period == SLAPeriod.DAILY:
            start = ref.replace(hour=0, minute=0, second=0, microsecond=0)
            end = start + timedelta(days=1) - timedelta(microseconds=1)
        
        elif period == SLAPeriod.WEEKLY:
            start = ref - timedelta(days=ref.weekday())
            start = start.replace(hour=0, minute=0, second=0, microsecond=0)
            end = start + timedelta(days=7) - timedelta(microseconds=1)
        
        elif period == SLAPeriod.MONTHLY:
            start = ref.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
            if ref.month == 12:
                end = start.replace(year=ref.year + 1, month=1)
            else:
                end = start.replace(month=ref.month + 1)
            end = end - timedelta(microseconds=1)
        
        elif period == SLAPeriod.QUARTERLY:
            quarter = (ref.month - 1) // 3
            start_month = quarter * 3 + 1
            start = ref.replace(month=start_month, day=1, hour=0, minute=0, second=0, microsecond=0)
            end_month = start_month + 3
            if end_month > 12:
                end = start.replace(year=ref.year + 1, month=end_month - 12)
            else:
                end = start.replace(month=end_month)
            end = end - timedelta(microseconds=1)
        
        elif period == SLAPeriod.YEARLY:
            start = ref.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
            end = start.replace(year=ref.year + 1) - timedelta(microseconds=1)
        
        else:
            start = ref.replace(hour=0, minute=0, second=0, microsecond=0)
            end = start + timedelta(days=1) - timedelta(microseconds=1)
        
        return start, end
    
    @staticmethod
    def get_period_minutes(period: SLAPeriod) -> float:
        """Get total minutes in period."""
        if period == SLAPeriod.HOURLY:
            return 60
        elif period == SLAPeriod.DAILY:
            return 24 * 60
        elif period == SLAPeriod.WEEKLY:
            return 7 * 24 * 60
        elif period == SLAPeriod.MONTHLY:
            return 30 * 24 * 60  # Approximate
        elif period == SLAPeriod.QUARTERLY:
            return 90 * 24 * 60
        elif period == SLAPeriod.YEARLY:
            return 365 * 24 * 60
        return 24 * 60


# SLA manager
class SLAManager:
    """SLA manager."""
    
    def __init__(
        self,
        sla_store: Optional[SLAStore] = None,
        metric_store: Optional[MetricStore] = None,
    ):
        self._sla_store = sla_store or InMemorySLAStore()
        self._metric_store = metric_store or InMemoryMetricStore()
        self._alert_handlers: List[Callable] = []
        self._alerts: List[SLAAlert] = []
        self._triggered_alerts: Set[str] = set()
    
    async def define(
        self,
        name: str,
        target: float,
        metric_type: Union[str, MetricType] = MetricType.UPTIME,
        period: Union[str, SLAPeriod] = SLAPeriod.MONTHLY,
        warning_threshold: Optional[float] = None,
        higher_is_better: bool = True,
        service: str = "",
        owner: str = "",
        **metadata,
    ) -> SLADefinition:
        """Define SLA."""
        if isinstance(metric_type, str):
            metric_type = MetricType(metric_type)
        if isinstance(period, str):
            period = SLAPeriod(period)
        
        if warning_threshold is None:
            # Default warning at 0.5% above target for uptime
            if metric_type == MetricType.UPTIME:
                warning_threshold = max(0, target - 0.4)
            else:
                warning_threshold = target * 0.95
        
        sla = SLADefinition(
            name=name,
            target=target,
            metric_type=metric_type,
            period=period,
            warning_threshold=warning_threshold,
            higher_is_better=higher_is_better,
            service=service,
            owner=owner,
            metadata=metadata,
        )
        
        await self._sla_store.save(sla)
        
        logger.info(f"SLA defined: {name} (target: {target}%)")
        
        return sla
    
    async def get(self, name: str) -> Optional[SLADefinition]:
        """Get SLA by name."""
        return await self._sla_store.get_by_name(name)
    
    async def list_slas(self) -> List[SLADefinition]:
        """List all SLAs."""
        return await self._sla_store.list_all()
    
    async def delete(self, name: str) -> bool:
        """Delete SLA."""
        sla = await self._sla_store.get_by_name(name)
        if sla:
            return await self._sla_store.delete(sla.id)
        return False
    
    async def record(
        self,
        sla_name: str,
        value: float,
        timestamp: Optional[datetime] = None,
        source: str = "",
        **metadata,
    ) -> Optional[MetricRecord]:
        """Record metric value."""
        sla = await self._sla_store.get_by_name(sla_name)
        
        if not sla:
            return None
        
        record = MetricRecord(
            sla_id=sla.id,
            value=value,
            timestamp=timestamp or datetime.utcnow(),
            source=source,
            metadata=metadata,
        )
        
        await self._metric_store.save(record)
        
        # Check for breaches
        await self._check_sla(sla)
        
        return record
    
    async def get_status(self, sla_name: str) -> Optional[SLAStatusReport]:
        """Get current SLA status."""
        sla = await self._sla_store.get_by_name(sla_name)
        
        if not sla:
            return None
        
        period_start, period_end = PeriodCalculator.get_period_range(sla.period)
        
        avg_value = await self._metric_store.get_average(
            sla.id,
            period_start,
            period_end,
        )
        
        records = await self._metric_store.list_by_sla(
            sla.id,
            period_start,
            period_end,
        )
        
        if avg_value is None:
            return SLAStatusReport(
                sla_id=sla.id,
                sla_name=sla.name,
                status=SLAStatus.UNKNOWN,
                target=sla.target,
                period_start=period_start,
                period_end=period_end,
            )
        
        # Determine status
        variance = avg_value - sla.target
        if not sla.higher_is_better:
            variance = -variance
        
        if sla.higher_is_better:
            is_compliant = avg_value >= sla.target
            is_at_risk = avg_value < sla.target and avg_value >= sla.warning_threshold
        else:
            is_compliant = avg_value <= sla.target
            is_at_risk = avg_value > sla.target and avg_value <= sla.warning_threshold
        
        if is_compliant:
            status = SLAStatus.COMPLIANT
        elif is_at_risk:
            status = SLAStatus.AT_RISK
        else:
            status = SLAStatus.BREACHED
        
        # Calculate error budget
        error_budget = await self.get_error_budget(sla_name)
        error_budget_remaining = error_budget.remaining_percentage if error_budget else 100.0
        
        return SLAStatusReport(
            sla_id=sla.id,
            sla_name=sla.name,
            status=status,
            current_value=avg_value,
            target=sla.target,
            variance=variance,
            period_start=period_start,
            period_end=period_end,
            records_count=len(records),
            is_breached=status == SLAStatus.BREACHED,
            error_budget_remaining=error_budget_remaining,
        )
    
    async def get_error_budget(self, sla_name: str) -> Optional[ErrorBudget]:
        """Calculate error budget."""
        sla = await self._sla_store.get_by_name(sla_name)
        
        if not sla:
            return None
        
        period_start, period_end = PeriodCalculator.get_period_range(sla.period)
        total_minutes = PeriodCalculator.get_period_minutes(sla.period)
        
        # Error budget = (100 - target) * total_minutes / 100
        allowed_downtime = (100 - sla.target) * total_minutes / 100
        
        # Get actual downtime from records
        records = await self._metric_store.list_by_sla(
            sla.id,
            period_start,
            period_end,
        )
        
        consumed = 0.0
        for record in records:
            if record.value < sla.target:
                # Simple calculation: assume each record represents 1 minute
                consumed += (sla.target - record.value) / 100
        
        remaining = max(0, allowed_downtime - consumed)
        remaining_pct = (remaining / allowed_downtime * 100) if allowed_downtime > 0 else 100
        
        return ErrorBudget(
            sla_id=sla.id,
            total_minutes=allowed_downtime,
            consumed_minutes=consumed,
            remaining_minutes=remaining,
            remaining_percentage=remaining_pct,
            period_start=period_start,
            period_end=period_end,
        )
    
    async def get_all_statuses(self) -> List[SLAStatusReport]:
        """Get status of all SLAs."""
        slas = await self._sla_store.list_all()
        statuses = []
        
        for sla in slas:
            if sla.enabled:
                status = await self.get_status(sla.name)
                if status:
                    statuses.append(status)
        
        return statuses
    
    async def get_breached_slas(self) -> List[SLAStatusReport]:
        """Get breached SLAs."""
        statuses = await self.get_all_statuses()
        return [s for s in statuses if s.status == SLAStatus.BREACHED]
    
    async def get_at_risk_slas(self) -> List[SLAStatusReport]:
        """Get at-risk SLAs."""
        statuses = await self.get_all_statuses()
        return [s for s in statuses if s.status == SLAStatus.AT_RISK]
    
    async def get_stats(self) -> SLAStats:
        """Get SLA statistics."""
        statuses = await self.get_all_statuses()
        
        stats = SLAStats(total_slas=len(statuses))
        
        compliance_values = []
        
        for status in statuses:
            if status.status == SLAStatus.COMPLIANT:
                stats.compliant += 1
            elif status.status == SLAStatus.AT_RISK:
                stats.at_risk += 1
            elif status.status == SLAStatus.BREACHED:
                stats.breached += 1
            
            if status.current_value > 0:
                compliance_values.append(status.current_value)
        
        if compliance_values:
            stats.avg_compliance = sum(compliance_values) / len(compliance_values)
        
        return stats
    
    def add_alert_handler(self, handler: Callable) -> None:
        """Add alert handler."""
        self._alert_handlers.append(handler)
    
    def get_alerts(self, include_resolved: bool = False) -> List[SLAAlert]:
        """Get alerts."""
        if include_resolved:
            return self._alerts.copy()
        return [a for a in self._alerts if not a.resolved]
    
    def acknowledge_alert(self, alert_id: str) -> bool:
        """Acknowledge alert."""
        for alert in self._alerts:
            if alert.id == alert_id:
                alert.acknowledged = True
                return True
        return False
    
    async def _check_sla(self, sla: SLADefinition) -> None:
        """Check SLA and trigger alerts."""
        status = await self.get_status(sla.name)
        
        if not status:
            return
        
        alert_key = f"{sla.id}:{status.status.value}"
        
        if status.status in (SLAStatus.AT_RISK, SLAStatus.BREACHED):
            if alert_key not in self._triggered_alerts:
                self._triggered_alerts.add(alert_key)
                
                severity = (
                    AlertSeverity.CRITICAL if status.status == SLAStatus.BREACHED
                    else AlertSeverity.WARNING
                )
                
                alert = SLAAlert(
                    sla_id=sla.id,
                    sla_name=sla.name,
                    severity=severity,
                    status=status.status,
                    current_value=status.current_value,
                    target=sla.target,
                    message=f"SLA '{sla.name}' is {status.status.value}: "
                            f"{status.current_value:.2f}% (target: {sla.target}%)",
                )
                
                self._alerts.append(alert)
                await self._notify_alert(alert)
        
        elif status.status == SLAStatus.COMPLIANT:
            # Clear triggered alerts for this SLA
            to_remove = [k for k in self._triggered_alerts if k.startswith(f"{sla.id}:")]
            for k in to_remove:
                self._triggered_alerts.discard(k)
    
    async def _notify_alert(self, alert: SLAAlert) -> None:
        """Notify alert handlers."""
        level = logging.WARNING if alert.severity == AlertSeverity.WARNING else logging.ERROR
        logger.log(level, f"SLA Alert: {alert.message}")
        
        for handler in self._alert_handlers:
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler(alert)
                else:
                    handler(alert)
            except Exception as e:
                logger.error(f"Alert handler error: {e}")


# Factory functions
def create_sla_manager() -> SLAManager:
    """Create SLA manager."""
    return SLAManager()


def create_sla_definition(
    name: str,
    target: float,
    **kwargs,
) -> SLADefinition:
    """Create SLA definition."""
    return SLADefinition(name=name, target=target, **kwargs)


__all__ = [
    # Exceptions
    "SLAError",
    "SLABreach",
    # Enums
    "MetricType",
    "SLAPeriod",
    "SLAStatus",
    "AlertSeverity",
    # Data classes
    "MetricRecord",
    "SLADefinition",
    "SLAStatusReport",
    "SLAAlert",
    "ErrorBudget",
    "SLAStats",
    # Stores
    "SLAStore",
    "InMemorySLAStore",
    "MetricStore",
    "InMemoryMetricStore",
    # Utilities
    "PeriodCalculator",
    # Manager
    "SLAManager",
    # Factory functions
    "create_sla_manager",
    "create_sla_definition",
]



from typing import Any, Dict
import logging
import threading

logger = logging.getLogger(__name__)


class ConfigurationManager:
    """Thread-safe configuration manager for framework components."""

    def __init__(self) -> None:
        self.configurations: Dict[str, Dict[str, Any]] = {}
        self._lock = threading.Lock()

    def set_config(self, component: str, config: Dict[str, Any]) -> None:
        with self._lock:
            self.configurations[component] = config
        logger.info("Configuration set for '%s'", component)

    def get_config(self, component: str) -> Dict[str, Any]:
        with self._lock:
            return self.configurations.get(component, {})

    def update_config(self, component: str, updates: Dict[str, Any]) -> None:
        with self._lock:
            if component in self.configurations:
                self.configurations[component].update(updates)
                logger.info("Configuration updated for '%s'", component)
            else:
                self.configurations[component] = updates
                logger.info("Configuration set for '%s'", component)

    def remove_config(self, component: str) -> None:
        with self._lock:
            if component in self.configurations:
                del self.configurations[component]
                logger.info("Configuration removed for '%s'", component)

    def list_components(self) -> list[str]:
        with self._lock:
            return list(self.configurations.keys())



"""
OpenAI Provider Adapter.

Provides seamless integration with OpenAI's API including:
- GPT-4, GPT-4o, GPT-3.5 models
- Function/tool calling
- Streaming
- Token counting with tiktoken
"""

import os
import logging
from typing import Any, Dict, Iterator, List, Optional

from .base import BaseLLMProvider, LLMMessage, LLMResponse, ProviderConfig

logger = logging.getLogger(__name__)


class OpenAIProvider(BaseLLMProvider):
    """
    OpenAI API provider adapter.
    
    Auto-configures from environment:
    - OPENAI_API_KEY
    - OPENAI_API_BASE (optional)
    - OPENAI_ORGANIZATION (optional)
    
    Example:
        >>> provider = OpenAIProvider.from_env()
        >>> response = provider.generate("Hello, world!")
        >>> print(response.content)
    """
    
    DEFAULT_MODEL = "gpt-4o"
    
    SUPPORTED_MODELS = [
        "gpt-4o",
        "gpt-4o-mini",
        "gpt-4-turbo",
        "gpt-4",
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-16k",
        "o1-preview",
        "o1-mini",
    ]
    
    def __init__(self, config: Optional[ProviderConfig] = None):
        super().__init__(config)
        if self.config.default_model is None:
            self.config.default_model = self.DEFAULT_MODEL
    
    @classmethod
    def from_env(cls, model: Optional[str] = None) -> 'OpenAIProvider':
        """Create provider from environment variables."""
        config = ProviderConfig(
            api_key=os.getenv("OPENAI_API_KEY"),
            api_base=os.getenv("OPENAI_API_BASE"),
            organization=os.getenv("OPENAI_ORGANIZATION"),
            default_model=model or os.getenv("OPENAI_MODEL", cls.DEFAULT_MODEL),
            timeout=float(os.getenv("OPENAI_TIMEOUT", "60")),
            max_retries=int(os.getenv("OPENAI_MAX_RETRIES", "3")),
        )
        return cls(config)
    
    @property
    def provider_name(self) -> str:
        return "openai"
    
    @property
    def supported_models(self) -> List[str]:
        return self.SUPPORTED_MODELS
    
    def _initialize_client(self) -> None:
        """Initialize OpenAI client."""
        try:
            from openai import OpenAI
            
            kwargs = {"api_key": self.config.api_key}
            if self.config.api_base:
                kwargs["base_url"] = self.config.api_base
            if self.config.organization:
                kwargs["organization"] = self.config.organization
            if self.config.timeout:
                kwargs["timeout"] = self.config.timeout
            if self.config.max_retries:
                kwargs["max_retries"] = self.config.max_retries
            
            self._client = OpenAI(**kwargs)
            logger.info("OpenAI client initialized")
        except ImportError:
            raise ImportError(
                "OpenAI package not installed. Install with: pip install openai"
            )
        except Exception as e:
            logger.error("Failed to initialize OpenAI client: %s", e)
            raise
    
    def generate(
        self,
        prompt: str,
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
        stop: Optional[List[str]] = None,
        **kwargs,
    ) -> LLMResponse:
        """Generate text using OpenAI API."""
        self._ensure_initialized()
        
        model = model or self.config.default_model
        
        messages = [{"role": "user", "content": prompt}]
        
        # Handle system prompt if provided
        if "system_prompt" in kwargs:
            messages.insert(0, {"role": "system", "content": kwargs.pop("system_prompt")})
        
        request_kwargs = {
            "model": model,
            "messages": messages,
            "temperature": temperature,
        }
        
        if max_tokens:
            request_kwargs["max_tokens"] = max_tokens
        if stop:
            request_kwargs["stop"] = stop
        
        # Pass through additional kwargs
        for key in ["top_p", "frequency_penalty", "presence_penalty", "seed"]:
            if key in kwargs:
                request_kwargs[key] = kwargs[key]
        
        try:
            response = self._client.chat.completions.create(**request_kwargs)
            
            return LLMResponse(
                content=response.choices[0].message.content or "",
                model=response.model,
                provider=self.provider_name,
                finish_reason=response.choices[0].finish_reason,
                usage={
                    "prompt_tokens": response.usage.prompt_tokens if response.usage else 0,
                    "completion_tokens": response.usage.completion_tokens if response.usage else 0,
                    "total_tokens": response.usage.total_tokens if response.usage else 0,
                },
                raw_response=response,
            )
        except Exception as e:
            logger.error("OpenAI generation failed: %s", e)
            raise
    
    def generate_chat(
        self,
        messages: List[LLMMessage],
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
        **kwargs,
    ) -> LLMResponse:
        """Generate from chat messages."""
        self._ensure_initialized()
        
        model = model or self.config.default_model
        
        # Convert to OpenAI format
        openai_messages = [m.to_dict() for m in messages]
        
        request_kwargs = {
            "model": model,
            "messages": openai_messages,
            "temperature": temperature,
        }
        
        if max_tokens:
            request_kwargs["max_tokens"] = max_tokens
        
        try:
            response = self._client.chat.completions.create(**request_kwargs)
            
            # Handle tool calls
            tool_calls = None
            message = response.choices[0].message
            if hasattr(message, 'tool_calls') and message.tool_calls:
                tool_calls = [
                    {
                        "id": tc.id,
                        "type": tc.type,
                        "function": {
                            "name": tc.function.name,
                            "arguments": tc.function.arguments,
                        }
                    }
                    for tc in message.tool_calls
                ]
            
            return LLMResponse(
                content=message.content or "",
                model=response.model,
                provider=self.provider_name,
                finish_reason=response.choices[0].finish_reason,
                tool_calls=tool_calls,
                usage={
                    "prompt_tokens": response.usage.prompt_tokens if response.usage else 0,
                    "completion_tokens": response.usage.completion_tokens if response.usage else 0,
                    "total_tokens": response.usage.total_tokens if response.usage else 0,
                },
                raw_response=response,
            )
        except Exception as e:
            logger.error("OpenAI chat generation failed: %s", e)
            raise
    
    def generate_with_tools(
        self,
        prompt: str,
        tools: List[Dict[str, Any]],
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        tool_choice: str = "auto",
        **kwargs,
    ) -> LLMResponse:
        """Generate with function/tool calling."""
        self._ensure_initialized()
        
        model = model or self.config.default_model
        
        messages = [{"role": "user", "content": prompt}]
        
        if "system_prompt" in kwargs:
            messages.insert(0, {"role": "system", "content": kwargs.pop("system_prompt")})
        
        # Convert tools to OpenAI format if needed
        openai_tools = []
        for tool in tools:
            if "type" not in tool:
                openai_tools.append({
                    "type": "function",
                    "function": tool,
                })
            else:
                openai_tools.append(tool)
        
        try:
            response = self._client.chat.completions.create(
                model=model,
                messages=messages,
                tools=openai_tools if openai_tools else None,
                tool_choice=tool_choice if openai_tools else None,
                temperature=temperature,
            )
            
            message = response.choices[0].message
            tool_calls = None
            
            if hasattr(message, 'tool_calls') and message.tool_calls:
                tool_calls = [
                    {
                        "id": tc.id,
                        "type": tc.type,
                        "function": {
                            "name": tc.function.name,
                            "arguments": tc.function.arguments,
                        }
                    }
                    for tc in message.tool_calls
                ]
            
            return LLMResponse(
                content=message.content or "",
                model=response.model,
                provider=self.provider_name,
                finish_reason=response.choices[0].finish_reason,
                tool_calls=tool_calls,
                usage={
                    "prompt_tokens": response.usage.prompt_tokens if response.usage else 0,
                    "completion_tokens": response.usage.completion_tokens if response.usage else 0,
                    "total_tokens": response.usage.total_tokens if response.usage else 0,
                },
                raw_response=response,
            )
        except Exception as e:
            logger.error("OpenAI tool generation failed: %s", e)
            raise
    
    def stream(
        self,
        prompt: str,
        *,
        model: Optional[str] = None,
        temperature: float = 0.7,
        **kwargs,
    ) -> Iterator[str]:
        """Stream tokens as they are generated."""
        self._ensure_initialized()
        
        model = model or self.config.default_model
        messages = [{"role": "user", "content": prompt}]
        
        if "system_prompt" in kwargs:
            messages.insert(0, {"role": "system", "content": kwargs.pop("system_prompt")})
        
        try:
            stream = self._client.chat.completions.create(
                model=model,
                messages=messages,
                temperature=temperature,
                stream=True,
            )
            
            for chunk in stream:
                if chunk.choices and chunk.choices[0].delta.content:
                    yield chunk.choices[0].delta.content
        except Exception as e:
            logger.error("OpenAI streaming failed: %s", e)
            raise
    
    def count_tokens(self, text: str, model: Optional[str] = None) -> int:
        """Count tokens using tiktoken."""
        model = model or self.config.default_model
        
        try:
            import tiktoken
            
            try:
                encoding = tiktoken.encoding_for_model(model)
            except KeyError:
                encoding = tiktoken.get_encoding("cl100k_base")
            
            return len(encoding.encode(text))
        except ImportError:
            # Fallback to rough estimate
            return len(text) // 4


__all__ = ['OpenAIProvider']



"""
Enterprise Cache Manager Module.

Provides cache management, eviction policies,
distributed caching, and cache-aside patterns.

Example:
    # Create cache
    cache = create_cache(ttl=300, max_size=1000)
    
    # Cache operations
    await cache.set("user:123", user_data)
    user = await cache.get("user:123")
    
    # Use decorator
    @cached(ttl=60)
    async def get_user(user_id: str):
        return await db.find_user(user_id)
"""

from __future__ import annotations

import asyncio
import hashlib
import json
import logging
import time
from abc import ABC, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
)

T = TypeVar('T')
K = TypeVar('K')
V = TypeVar('V')


logger = logging.getLogger(__name__)


class CacheError(Exception):
    """Cache error."""
    pass


class CacheMissError(CacheError):
    """Cache miss error."""
    pass


class CacheFullError(CacheError):
    """Cache full error."""
    pass


class EvictionPolicy(str, Enum):
    """Cache eviction policy."""
    LRU = "lru"  # Least Recently Used
    LFU = "lfu"  # Least Frequently Used
    FIFO = "fifo"  # First In First Out
    TTL = "ttl"  # Time To Live
    RANDOM = "random"


class WritePolicy(str, Enum):
    """Cache write policy."""
    WRITE_THROUGH = "write_through"
    WRITE_BACK = "write_back"
    WRITE_AROUND = "write_around"


@dataclass
class CacheEntry(Generic[V]):
    """Cache entry with metadata."""
    key: str
    value: V
    created_at: datetime = field(default_factory=datetime.now)
    accessed_at: datetime = field(default_factory=datetime.now)
    expires_at: Optional[datetime] = None
    access_count: int = 0
    size_bytes: int = 0
    tags: Set[str] = field(default_factory=set)
    
    @property
    def is_expired(self) -> bool:
        if self.expires_at is None:
            return False
        return datetime.now() > self.expires_at
    
    def touch(self) -> None:
        """Update access time and count."""
        self.accessed_at = datetime.now()
        self.access_count += 1


@dataclass
class CacheConfig:
    """Cache configuration."""
    max_size: int = 1000
    ttl: Optional[float] = None  # Default TTL in seconds
    eviction_policy: EvictionPolicy = EvictionPolicy.LRU
    write_policy: WritePolicy = WritePolicy.WRITE_THROUGH
    namespace: str = ""


@dataclass
class CacheStats:
    """Cache statistics."""
    hits: int = 0
    misses: int = 0
    sets: int = 0
    deletes: int = 0
    evictions: int = 0
    expirations: int = 0
    size: int = 0
    max_size: int = 0
    
    @property
    def hit_ratio(self) -> float:
        total = self.hits + self.misses
        return self.hits / total if total > 0 else 0.0


class Serializer(ABC, Generic[T]):
    """
    Cache value serializer.
    """
    
    @abstractmethod
    def serialize(self, value: T) -> bytes:
        """Serialize value."""
        pass
    
    @abstractmethod
    def deserialize(self, data: bytes) -> T:
        """Deserialize value."""
        pass


class JsonSerializer(Serializer[Any]):
    """JSON serializer."""
    
    def serialize(self, value: Any) -> bytes:
        return json.dumps(value, default=str).encode('utf-8')
    
    def deserialize(self, data: bytes) -> Any:
        return json.loads(data.decode('utf-8'))


class PickleSerializer(Serializer[Any]):
    """Pickle serializer."""
    
    def serialize(self, value: Any) -> bytes:
        import pickle
        return pickle.dumps(value)
    
    def deserialize(self, data: bytes) -> Any:
        import pickle
        return pickle.loads(data)


class Cache(ABC, Generic[K, V]):
    """
    Abstract cache interface.
    """
    
    @abstractmethod
    async def get(self, key: K) -> Optional[V]:
        """Get value from cache."""
        pass
    
    @abstractmethod
    async def set(
        self,
        key: K,
        value: V,
        ttl: Optional[float] = None,
        tags: Optional[Set[str]] = None,
    ) -> None:
        """Set value in cache."""
        pass
    
    @abstractmethod
    async def delete(self, key: K) -> bool:
        """Delete value from cache."""
        pass
    
    @abstractmethod
    async def exists(self, key: K) -> bool:
        """Check if key exists."""
        pass
    
    @abstractmethod
    async def clear(self) -> None:
        """Clear all entries."""
        pass
    
    @abstractmethod
    async def stats(self) -> CacheStats:
        """Get cache statistics."""
        pass


class InMemoryCache(Cache[str, Any]):
    """
    In-memory cache implementation.
    """
    
    def __init__(self, config: Optional[CacheConfig] = None):
        self._config = config or CacheConfig()
        self._entries: OrderedDict[str, CacheEntry] = OrderedDict()
        self._stats = CacheStats(max_size=self._config.max_size)
        self._lock = asyncio.Lock()
    
    def _make_key(self, key: str) -> str:
        """Create namespaced key."""
        if self._config.namespace:
            return f"{self._config.namespace}:{key}"
        return key
    
    async def get(self, key: str) -> Optional[Any]:
        async with self._lock:
            full_key = self._make_key(key)
            entry = self._entries.get(full_key)
            
            if entry is None:
                self._stats.misses += 1
                return None
            
            if entry.is_expired:
                del self._entries[full_key]
                self._stats.expirations += 1
                self._stats.misses += 1
                self._stats.size -= 1
                return None
            
            entry.touch()
            
            # Move to end for LRU
            if self._config.eviction_policy == EvictionPolicy.LRU:
                self._entries.move_to_end(full_key)
            
            self._stats.hits += 1
            return entry.value
    
    async def set(
        self,
        key: str,
        value: Any,
        ttl: Optional[float] = None,
        tags: Optional[Set[str]] = None,
    ) -> None:
        async with self._lock:
            full_key = self._make_key(key)
            
            # Evict if at capacity
            while len(self._entries) >= self._config.max_size:
                await self._evict()
            
            expires_at = None
            effective_ttl = ttl or self._config.ttl
            if effective_ttl:
                expires_at = datetime.now() + timedelta(seconds=effective_ttl)
            
            entry = CacheEntry(
                key=full_key,
                value=value,
                expires_at=expires_at,
                tags=tags or set(),
            )
            
            self._entries[full_key] = entry
            self._stats.sets += 1
            self._stats.size = len(self._entries)
    
    async def delete(self, key: str) -> bool:
        async with self._lock:
            full_key = self._make_key(key)
            if full_key in self._entries:
                del self._entries[full_key]
                self._stats.deletes += 1
                self._stats.size = len(self._entries)
                return True
            return False
    
    async def exists(self, key: str) -> bool:
        async with self._lock:
            full_key = self._make_key(key)
            entry = self._entries.get(full_key)
            
            if entry is None:
                return False
            
            if entry.is_expired:
                del self._entries[full_key]
                self._stats.expirations += 1
                self._stats.size -= 1
                return False
            
            return True
    
    async def clear(self) -> None:
        async with self._lock:
            self._entries.clear()
            self._stats.size = 0
    
    async def stats(self) -> CacheStats:
        return self._stats
    
    async def _evict(self) -> None:
        """Evict an entry based on policy."""
        if not self._entries:
            return
        
        if self._config.eviction_policy == EvictionPolicy.LRU:
            # Remove least recently used (first item)
            self._entries.popitem(last=False)
        elif self._config.eviction_policy == EvictionPolicy.LFU:
            # Remove least frequently used
            min_entry = min(
                self._entries.items(),
                key=lambda x: x[1].access_count,
            )
            del self._entries[min_entry[0]]
        elif self._config.eviction_policy == EvictionPolicy.FIFO:
            # Remove first inserted
            self._entries.popitem(last=False)
        elif self._config.eviction_policy == EvictionPolicy.RANDOM:
            import random
            key = random.choice(list(self._entries.keys()))
            del self._entries[key]
        else:
            self._entries.popitem(last=False)
        
        self._stats.evictions += 1
        self._stats.size = len(self._entries)
    
    async def get_by_tag(self, tag: str) -> List[Tuple[str, Any]]:
        """Get all entries with a tag."""
        async with self._lock:
            return [
                (e.key, e.value)
                for e in self._entries.values()
                if tag in e.tags and not e.is_expired
            ]
    
    async def delete_by_tag(self, tag: str) -> int:
        """Delete all entries with a tag."""
        async with self._lock:
            keys_to_delete = [
                key for key, entry in self._entries.items()
                if tag in entry.tags
            ]
            for key in keys_to_delete:
                del self._entries[key]
            
            self._stats.deletes += len(keys_to_delete)
            self._stats.size = len(self._entries)
            return len(keys_to_delete)


class TieredCache(Cache[str, Any]):
    """
    Tiered cache with multiple levels.
    """
    
    def __init__(self, caches: List[Cache]):
        self._caches = caches
    
    async def get(self, key: str) -> Optional[Any]:
        for i, cache in enumerate(self._caches):
            value = await cache.get(key)
            if value is not None:
                # Populate upper tiers
                for j in range(i):
                    await self._caches[j].set(key, value)
                return value
        return None
    
    async def set(
        self,
        key: str,
        value: Any,
        ttl: Optional[float] = None,
        tags: Optional[Set[str]] = None,
    ) -> None:
        for cache in self._caches:
            await cache.set(key, value, ttl, tags)
    
    async def delete(self, key: str) -> bool:
        deleted = False
        for cache in self._caches:
            if await cache.delete(key):
                deleted = True
        return deleted
    
    async def exists(self, key: str) -> bool:
        for cache in self._caches:
            if await cache.exists(key):
                return True
        return False
    
    async def clear(self) -> None:
        for cache in self._caches:
            await cache.clear()
    
    async def stats(self) -> CacheStats:
        combined = CacheStats()
        for cache in self._caches:
            s = await cache.stats()
            combined.hits += s.hits
            combined.misses += s.misses
            combined.sets += s.sets
            combined.deletes += s.deletes
            combined.evictions += s.evictions
            combined.size += s.size
        return combined


class CacheAside:
    """
    Cache-aside pattern implementation.
    """
    
    def __init__(
        self,
        cache: Cache,
        loader: Callable[[str], Any],
        ttl: Optional[float] = None,
    ):
        self._cache = cache
        self._loader = loader
        self._ttl = ttl
    
    async def get(self, key: str) -> Any:
        """Get from cache or load."""
        value = await self._cache.get(key)
        
        if value is not None:
            return value
        
        # Load from source
        if asyncio.iscoroutinefunction(self._loader):
            value = await self._loader(key)
        else:
            value = self._loader(key)
        
        if value is not None:
            await self._cache.set(key, value, ttl=self._ttl)
        
        return value
    
    async def invalidate(self, key: str) -> None:
        """Invalidate cache entry."""
        await self._cache.delete(key)
    
    async def refresh(self, key: str) -> Any:
        """Refresh cache entry."""
        await self._cache.delete(key)
        return await self.get(key)


class WriteThrough:
    """
    Write-through cache pattern.
    """
    
    def __init__(
        self,
        cache: Cache,
        writer: Callable[[str, Any], Any],
        ttl: Optional[float] = None,
    ):
        self._cache = cache
        self._writer = writer
        self._ttl = ttl
    
    async def write(self, key: str, value: Any) -> None:
        """Write to cache and source."""
        # Write to source first
        if asyncio.iscoroutinefunction(self._writer):
            await self._writer(key, value)
        else:
            self._writer(key, value)
        
        # Then update cache
        await self._cache.set(key, value, ttl=self._ttl)


class CacheManager:
    """
    Cache manager for multiple caches.
    """
    
    def __init__(self):
        self._caches: Dict[str, Cache] = {}
        self._default: Optional[str] = None
    
    def register(
        self,
        name: str,
        cache: Cache,
        default: bool = False,
    ) -> None:
        """Register a cache."""
        self._caches[name] = cache
        if default or self._default is None:
            self._default = name
    
    def get_cache(self, name: Optional[str] = None) -> Cache:
        """Get a cache."""
        name = name or self._default
        if not name or name not in self._caches:
            raise CacheError(f"Cache not found: {name}")
        return self._caches[name]
    
    async def clear_all(self) -> None:
        """Clear all caches."""
        for cache in self._caches.values():
            await cache.clear()
    
    async def stats_all(self) -> Dict[str, CacheStats]:
        """Get stats for all caches."""
        return {
            name: await cache.stats()
            for name, cache in self._caches.items()
        }


# Global manager
_global_manager = CacheManager()


def _generate_cache_key(
    func: Callable,
    args: tuple,
    kwargs: dict,
    key_builder: Optional[Callable] = None,
) -> str:
    """Generate cache key for function call."""
    if key_builder:
        return key_builder(*args, **kwargs)
    
    # Default key generation
    key_parts = [func.__module__, func.__qualname__]
    
    # Add args
    for arg in args:
        key_parts.append(str(hash(str(arg))))
    
    # Add kwargs
    for k, v in sorted(kwargs.items()):
        key_parts.append(f"{k}={hash(str(v))}")
    
    key_str = ":".join(key_parts)
    return hashlib.md5(key_str.encode()).hexdigest()


# Decorators
def cached(
    ttl: Optional[float] = None,
    cache_name: Optional[str] = None,
    key_builder: Optional[Callable[..., str]] = None,
    tags: Optional[Set[str]] = None,
) -> Callable:
    """
    Decorator to cache function results.
    
    Example:
        @cached(ttl=60)
        async def get_user(user_id: str):
            return await db.find_user(user_id)
    """
    def decorator(func: Callable) -> Callable:
        async def wrapper(*args, **kwargs):
            cache = _global_manager.get_cache(cache_name)
            key = _generate_cache_key(func, args, kwargs, key_builder)
            
            # Try cache
            value = await cache.get(key)
            if value is not None:
                return value
            
            # Execute function
            if asyncio.iscoroutinefunction(func):
                value = await func(*args, **kwargs)
            else:
                value = func(*args, **kwargs)
            
            # Cache result
            if value is not None:
                await cache.set(key, value, ttl=ttl, tags=tags)
            
            return value
        
        wrapper._cached = True
        wrapper._cache_name = cache_name
        return wrapper
    
    return decorator


def cache_invalidate(
    key_builder: Callable[..., str],
    cache_name: Optional[str] = None,
) -> Callable:
    """
    Decorator to invalidate cache on function call.
    
    Example:
        @cache_invalidate(lambda user_id: f"user:{user_id}")
        async def update_user(user_id: str, data: dict):
            ...
    """
    def decorator(func: Callable) -> Callable:
        async def wrapper(*args, **kwargs):
            # Execute function first
            if asyncio.iscoroutinefunction(func):
                result = await func(*args, **kwargs)
            else:
                result = func(*args, **kwargs)
            
            # Invalidate cache
            cache = _global_manager.get_cache(cache_name)
            key = key_builder(*args, **kwargs)
            await cache.delete(key)
            
            return result
        
        return wrapper
    
    return decorator


# Factory functions
def create_cache(
    max_size: int = 1000,
    ttl: Optional[float] = None,
    eviction_policy: EvictionPolicy = EvictionPolicy.LRU,
    namespace: str = "",
) -> InMemoryCache:
    """Create an in-memory cache."""
    return InMemoryCache(CacheConfig(
        max_size=max_size,
        ttl=ttl,
        eviction_policy=eviction_policy,
        namespace=namespace,
    ))


def create_lru_cache(
    max_size: int = 1000,
    ttl: Optional[float] = None,
) -> InMemoryCache:
    """Create an LRU cache."""
    return create_cache(
        max_size=max_size,
        ttl=ttl,
        eviction_policy=EvictionPolicy.LRU,
    )


def create_lfu_cache(
    max_size: int = 1000,
    ttl: Optional[float] = None,
) -> InMemoryCache:
    """Create an LFU cache."""
    return create_cache(
        max_size=max_size,
        ttl=ttl,
        eviction_policy=EvictionPolicy.LFU,
    )


def create_tiered_cache(
    *caches: Cache,
) -> TieredCache:
    """Create a tiered cache."""
    return TieredCache(list(caches))


def create_cache_aside(
    cache: Cache,
    loader: Callable[[str], Any],
    ttl: Optional[float] = None,
) -> CacheAside:
    """Create a cache-aside pattern."""
    return CacheAside(cache, loader, ttl)


def create_write_through(
    cache: Cache,
    writer: Callable[[str, Any], Any],
    ttl: Optional[float] = None,
) -> WriteThrough:
    """Create a write-through pattern."""
    return WriteThrough(cache, writer, ttl)


def register_cache(
    name: str,
    cache: Cache,
    default: bool = False,
) -> None:
    """Register cache in global manager."""
    _global_manager.register(name, cache, default)


def get_cache(name: Optional[str] = None) -> Cache:
    """Get cache from global manager."""
    return _global_manager.get_cache(name)


__all__ = [
    # Exceptions
    "CacheError",
    "CacheMissError",
    "CacheFullError",
    # Enums
    "EvictionPolicy",
    "WritePolicy",
    # Data classes
    "CacheEntry",
    "CacheConfig",
    "CacheStats",
    # Serializers
    "Serializer",
    "JsonSerializer",
    "PickleSerializer",
    # Cache
    "Cache",
    "InMemoryCache",
    "TieredCache",
    # Patterns
    "CacheAside",
    "WriteThrough",
    # Manager
    "CacheManager",
    # Decorators
    "cached",
    "cache_invalidate",
    # Factory functions
    "create_cache",
    "create_lru_cache",
    "create_lfu_cache",
    "create_tiered_cache",
    "create_cache_aside",
    "create_write_through",
    "register_cache",
    "get_cache",
]



"""
Enterprise RAG (Retrieval-Augmented Generation) Module.

Provides retrieval-augmented generation patterns, context injection,
document retrieval, and answer generation for agents.

Example:
    # Create RAG pipeline
    rag = RAGPipeline(
        retriever=VectorRetriever(store=vector_store),
        generator=LLMGenerator(client=openai_client),
    )
    
    # Query
    answer = await rag.query("What is RAG?")
    
    # With context builder
    context = await rag.retrieve("What is RAG?", k=5)
    response = await rag.generate(query, context)
"""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
from datetime import datetime
from functools import wraps
from enum import Enum
import logging
import time

logger = logging.getLogger(__name__)

T = TypeVar('T')


class RAGError(Exception):
    """RAG pipeline error."""
    pass


class RetrievalError(RAGError):
    """Document retrieval error."""
    pass


class GenerationError(RAGError):
    """Response generation error."""
    pass


class RetrievalStrategy(str, Enum):
    """Retrieval strategies."""
    DENSE = "dense"           # Vector similarity
    SPARSE = "sparse"         # BM25/keyword
    HYBRID = "hybrid"         # Dense + sparse
    RERANK = "rerank"         # With reranking


class ChunkStrategy(str, Enum):
    """Document chunking strategies."""
    FIXED = "fixed"           # Fixed token count
    SENTENCE = "sentence"     # Sentence boundaries
    PARAGRAPH = "paragraph"   # Paragraph boundaries
    SEMANTIC = "semantic"     # Semantic similarity


@dataclass
class Document:
    """A retrievable document."""
    id: str
    content: str
    metadata: Dict[str, Any] = field(default_factory=dict)
    score: float = 0.0
    source: Optional[str] = None
    chunk_index: int = 0
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "content": self.content,
            "metadata": self.metadata,
            "score": self.score,
            "source": self.source,
        }


@dataclass
class RetrievalResult:
    """Result of document retrieval."""
    documents: List[Document]
    query: str
    strategy: RetrievalStrategy = RetrievalStrategy.DENSE
    duration_ms: float = 0.0
    
    @property
    def contents(self) -> List[str]:
        """Get document contents."""
        return [doc.content for doc in self.documents]
    
    @property
    def context(self) -> str:
        """Get combined context."""
        return "\n\n".join(self.contents)
    
    def __len__(self) -> int:
        return len(self.documents)


@dataclass
class RAGResponse:
    """RAG response with sources."""
    answer: str
    query: str
    sources: List[Document]
    context_used: str = ""
    model: str = ""
    duration_ms: float = 0.0
    tokens_used: int = 0
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "answer": self.answer,
            "query": self.query,
            "sources": [doc.to_dict() for doc in self.sources],
            "model": self.model,
            "duration_ms": self.duration_ms,
        }


class Retriever(ABC):
    """Abstract retriever interface."""
    
    @abstractmethod
    async def retrieve(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict[str, Any]] = None,
    ) -> List[Document]:
        """Retrieve relevant documents."""
        pass


class VectorRetriever(Retriever):
    """Vector-based retriever using embeddings."""
    
    def __init__(
        self,
        store: Any,  # VectorStore
        embedder: Optional[Any] = None,
    ):
        self._store = store
        self._embedder = embedder
    
    async def retrieve(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict[str, Any]] = None,
    ) -> List[Document]:
        """Retrieve using vector similarity."""
        results = await self._store.search(query, k=k, filter=filter)
        
        return [
            Document(
                id=r.id if hasattr(r, 'id') else str(i),
                content=r.text if hasattr(r, 'text') else str(r),
                score=r.score if hasattr(r, 'score') else 0.0,
                metadata=r.metadata if hasattr(r, 'metadata') else {},
            )
            for i, r in enumerate(results)
        ]


class KeywordRetriever(Retriever):
    """Keyword-based retriever (BM25-like)."""
    
    def __init__(self, documents: List[Document]):
        self._documents = documents
        self._index: Dict[str, List[int]] = {}
        self._build_index()
    
    def _build_index(self) -> None:
        """Build inverted index."""
        for i, doc in enumerate(self._documents):
            words = doc.content.lower().split()
            for word in set(words):
                if word not in self._index:
                    self._index[word] = []
                self._index[word].append(i)
    
    async def retrieve(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict[str, Any]] = None,
    ) -> List[Document]:
        """Retrieve using keyword matching."""
        words = query.lower().split()
        scores: Dict[int, float] = {}
        
        for word in words:
            if word in self._index:
                for doc_idx in self._index[word]:
                    scores[doc_idx] = scores.get(doc_idx, 0) + 1
        
        # Sort by score
        sorted_indices = sorted(scores.keys(), key=lambda x: scores[x], reverse=True)
        
        results = []
        for idx in sorted_indices[:k]:
            doc = self._documents[idx]
            doc.score = scores[idx]
            
            # Apply filter
            if filter:
                if not all(doc.metadata.get(key) == value for key, value in filter.items()):
                    continue
            
            results.append(doc)
        
        return results[:k]


class HybridRetriever(Retriever):
    """Hybrid retriever combining dense and sparse."""
    
    def __init__(
        self,
        dense_retriever: Retriever,
        sparse_retriever: Retriever,
        dense_weight: float = 0.7,
    ):
        self._dense = dense_retriever
        self._sparse = sparse_retriever
        self._dense_weight = dense_weight
    
    async def retrieve(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict[str, Any]] = None,
    ) -> List[Document]:
        """Retrieve using hybrid approach."""
        # Retrieve from both
        dense_results, sparse_results = await asyncio.gather(
            self._dense.retrieve(query, k=k * 2, filter=filter),
            self._sparse.retrieve(query, k=k * 2, filter=filter),
        )
        
        # Combine scores using Reciprocal Rank Fusion
        scores: Dict[str, float] = {}
        docs: Dict[str, Document] = {}
        
        for rank, doc in enumerate(dense_results):
            rrf_score = self._dense_weight / (rank + 60)
            scores[doc.id] = scores.get(doc.id, 0) + rrf_score
            docs[doc.id] = doc
        
        sparse_weight = 1 - self._dense_weight
        for rank, doc in enumerate(sparse_results):
            rrf_score = sparse_weight / (rank + 60)
            scores[doc.id] = scores.get(doc.id, 0) + rrf_score
            if doc.id not in docs:
                docs[doc.id] = doc
        
        # Sort by combined score
        sorted_ids = sorted(scores.keys(), key=lambda x: scores[x], reverse=True)
        
        results = []
        for doc_id in sorted_ids[:k]:
            doc = docs[doc_id]
            doc.score = scores[doc_id]
            results.append(doc)
        
        return results


class RerankingRetriever(Retriever):
    """Retriever with reranking."""
    
    def __init__(
        self,
        base_retriever: Retriever,
        reranker: Callable[[str, List[Document]], List[Document]],
        initial_k: int = 20,
    ):
        self._base = base_retriever
        self._reranker = reranker
        self._initial_k = initial_k
    
    async def retrieve(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict[str, Any]] = None,
    ) -> List[Document]:
        """Retrieve and rerank."""
        # Get more candidates
        candidates = await self._base.retrieve(
            query, k=self._initial_k, filter=filter
        )
        
        # Rerank
        if asyncio.iscoroutinefunction(self._reranker):
            reranked = await self._reranker(query, candidates)
        else:
            reranked = self._reranker(query, candidates)
        
        return reranked[:k]


class Generator(ABC):
    """Abstract generator interface."""
    
    @abstractmethod
    async def generate(
        self,
        query: str,
        context: str,
        **kwargs: Any,
    ) -> str:
        """Generate response given query and context."""
        pass


class LLMGenerator(Generator):
    """LLM-based response generator."""
    
    DEFAULT_TEMPLATE = """Answer the question based on the following context.
If the answer cannot be found in the context, say "I don't have enough information to answer this question."

Context:
{context}

Question: {query}

Answer:"""
    
    def __init__(
        self,
        client: Any,
        model: Optional[str] = None,
        template: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
    ):
        self._client = client
        self._model = model
        self._template = template or self.DEFAULT_TEMPLATE
        self._temperature = temperature
        self._max_tokens = max_tokens
    
    async def generate(
        self,
        query: str,
        context: str,
        **kwargs: Any,
    ) -> str:
        """Generate response using LLM."""
        prompt = self._template.format(query=query, context=context)
        
        # Handle different client interfaces
        if hasattr(self._client, 'chat'):
            response = await self._client.chat.completions.create(
                model=self._model,
                messages=[{"role": "user", "content": prompt}],
                temperature=kwargs.get("temperature", self._temperature),
                max_tokens=kwargs.get("max_tokens", self._max_tokens),
            )
            return response.choices[0].message.content
        
        if hasattr(self._client, 'generate'):
            return await self._client.generate(prompt=prompt, **kwargs)
        
        raise GenerationError(f"Unsupported client: {type(self._client)}")


class ContextBuilder:
    """Builds context from retrieved documents."""
    
    def __init__(
        self,
        max_tokens: int = 4000,
        token_counter: Optional[Callable[[str], int]] = None,
        include_metadata: bool = False,
        separator: str = "\n\n---\n\n",
    ):
        self._max_tokens = max_tokens
        self._token_counter = token_counter or (lambda s: len(s) // 4)
        self._include_metadata = include_metadata
        self._separator = separator
    
    def build(self, documents: List[Document]) -> str:
        """Build context from documents."""
        parts = []
        total_tokens = 0
        
        for i, doc in enumerate(documents):
            part = self._format_document(doc, i + 1)
            tokens = self._token_counter(part)
            
            if total_tokens + tokens > self._max_tokens:
                break
            
            parts.append(part)
            total_tokens += tokens
        
        return self._separator.join(parts)
    
    def _format_document(self, doc: Document, index: int) -> str:
        """Format a single document."""
        parts = [f"[{index}] {doc.content}"]
        
        if self._include_metadata and doc.metadata:
            meta_str = ", ".join(f"{k}: {v}" for k, v in doc.metadata.items())
            parts.append(f"Metadata: {meta_str}")
        
        if doc.source:
            parts.append(f"Source: {doc.source}")
        
        return "\n".join(parts)


class RAGPipeline:
    """
    Complete RAG pipeline.
    """
    
    def __init__(
        self,
        retriever: Retriever,
        generator: Generator,
        context_builder: Optional[ContextBuilder] = None,
        pre_processors: Optional[List[Callable[[str], str]]] = None,
        post_processors: Optional[List[Callable[[str], str]]] = None,
    ):
        self._retriever = retriever
        self._generator = generator
        self._context_builder = context_builder or ContextBuilder()
        self._pre_processors = pre_processors or []
        self._post_processors = post_processors or []
    
    async def query(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> RAGResponse:
        """
        Execute RAG query.
        
        Args:
            query: User query
            k: Number of documents to retrieve
            filter: Metadata filter
            **kwargs: Additional generation kwargs
        """
        start = time.time()
        
        # Pre-process query
        processed_query = query
        for processor in self._pre_processors:
            processed_query = processor(processed_query)
        
        # Retrieve
        documents = await self._retriever.retrieve(
            processed_query, k=k, filter=filter
        )
        
        # Build context
        context = self._context_builder.build(documents)
        
        # Generate
        answer = await self._generator.generate(
            processed_query, context, **kwargs
        )
        
        # Post-process
        for processor in self._post_processors:
            answer = processor(answer)
        
        duration = (time.time() - start) * 1000
        
        return RAGResponse(
            answer=answer,
            query=query,
            sources=documents,
            context_used=context,
            duration_ms=duration,
        )
    
    async def retrieve(
        self,
        query: str,
        k: int = 5,
        filter: Optional[Dict[str, Any]] = None,
    ) -> RetrievalResult:
        """Retrieve documents only."""
        start = time.time()
        
        documents = await self._retriever.retrieve(query, k=k, filter=filter)
        
        return RetrievalResult(
            documents=documents,
            query=query,
            duration_ms=(time.time() - start) * 1000,
        )
    
    async def generate(
        self,
        query: str,
        context: str,
        **kwargs: Any,
    ) -> str:
        """Generate response only."""
        return await self._generator.generate(query, context, **kwargs)


class DocumentChunker:
    """Chunk documents for indexing."""
    
    def __init__(
        self,
        strategy: ChunkStrategy = ChunkStrategy.FIXED,
        chunk_size: int = 512,
        overlap: int = 50,
    ):
        self._strategy = strategy
        self._chunk_size = chunk_size
        self._overlap = overlap
    
    def chunk(self, document: Document) -> List[Document]:
        """Chunk a document."""
        if self._strategy == ChunkStrategy.FIXED:
            return self._chunk_fixed(document)
        elif self._strategy == ChunkStrategy.SENTENCE:
            return self._chunk_sentence(document)
        elif self._strategy == ChunkStrategy.PARAGRAPH:
            return self._chunk_paragraph(document)
        else:
            return [document]
    
    def _chunk_fixed(self, document: Document) -> List[Document]:
        """Chunk by fixed token count."""
        words = document.content.split()
        chunks = []
        start = 0
        
        while start < len(words):
            end = start + self._chunk_size
            chunk_words = words[start:end]
            
            chunks.append(Document(
                id=f"{document.id}_chunk_{len(chunks)}",
                content=" ".join(chunk_words),
                metadata={**document.metadata, "parent_id": document.id},
                source=document.source,
                chunk_index=len(chunks),
            ))
            
            start = end - self._overlap
        
        return chunks
    
    def _chunk_sentence(self, document: Document) -> List[Document]:
        """Chunk by sentence boundaries."""
        import re
        sentences = re.split(r'(?<=[.!?])\s+', document.content)
        
        chunks = []
        current_chunk = []
        current_size = 0
        
        for sentence in sentences:
            words = len(sentence.split())
            
            if current_size + words > self._chunk_size and current_chunk:
                chunks.append(Document(
                    id=f"{document.id}_chunk_{len(chunks)}",
                    content=" ".join(current_chunk),
                    metadata={**document.metadata, "parent_id": document.id},
                    source=document.source,
                    chunk_index=len(chunks),
                ))
                current_chunk = []
                current_size = 0
            
            current_chunk.append(sentence)
            current_size += words
        
        if current_chunk:
            chunks.append(Document(
                id=f"{document.id}_chunk_{len(chunks)}",
                content=" ".join(current_chunk),
                metadata={**document.metadata, "parent_id": document.id},
                source=document.source,
                chunk_index=len(chunks),
            ))
        
        return chunks
    
    def _chunk_paragraph(self, document: Document) -> List[Document]:
        """Chunk by paragraph boundaries."""
        paragraphs = document.content.split("\n\n")
        
        chunks = []
        for i, para in enumerate(paragraphs):
            if para.strip():
                chunks.append(Document(
                    id=f"{document.id}_chunk_{i}",
                    content=para.strip(),
                    metadata={**document.metadata, "parent_id": document.id},
                    source=document.source,
                    chunk_index=i,
                ))
        
        return chunks


class ConversationalRAG:
    """
    RAG with conversation history.
    """
    
    def __init__(
        self,
        rag_pipeline: RAGPipeline,
        max_history: int = 5,
    ):
        self._rag = rag_pipeline
        self._max_history = max_history
        self._history: List[Tuple[str, str]] = []
    
    async def query(
        self,
        query: str,
        k: int = 5,
        **kwargs: Any,
    ) -> RAGResponse:
        """Query with conversation context."""
        # Build conversation-aware query
        if self._history:
            history_context = "\n".join(
                f"User: {q}\nAssistant: {a}"
                for q, a in self._history[-self._max_history:]
            )
            enhanced_query = f"Given this conversation:\n{history_context}\n\nNew question: {query}"
        else:
            enhanced_query = query
        
        # Execute RAG
        response = await self._rag.query(enhanced_query, k=k, **kwargs)
        
        # Update history
        self._history.append((query, response.answer))
        
        return response
    
    def clear_history(self) -> None:
        """Clear conversation history."""
        self._history.clear()


# Decorators
def with_rag(
    rag_pipeline: RAGPipeline,
    k: int = 5,
) -> Callable:
    """
    Decorator to enhance function with RAG.
    
    Example:
        @with_rag(rag_pipeline)
        async def answer_question(query: str) -> str:
            ...
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(query: str, *args: Any, **kwargs: Any) -> RAGResponse:
            return await rag_pipeline.query(query, k=k, **kwargs)
        
        return wrapper
    
    return decorator


# Factory functions
def create_rag_pipeline(
    vector_store: Any,
    llm_client: Any,
    model: Optional[str] = None,
    retrieval_k: int = 5,
    max_context_tokens: int = 4000,
    **kwargs: Any,
) -> RAGPipeline:
    """
    Factory function to create a RAG pipeline.
    
    Args:
        vector_store: Vector store for retrieval
        llm_client: LLM client for generation
        model: Model name
        retrieval_k: Default number of documents
        max_context_tokens: Max context size
    """
    retriever = VectorRetriever(store=vector_store)
    generator = LLMGenerator(client=llm_client, model=model, **kwargs)
    context_builder = ContextBuilder(max_tokens=max_context_tokens)
    
    return RAGPipeline(
        retriever=retriever,
        generator=generator,
        context_builder=context_builder,
    )


__all__ = [
    # Exceptions
    "RAGError",
    "RetrievalError",
    "GenerationError",
    # Enums
    "RetrievalStrategy",
    "ChunkStrategy",
    # Data classes
    "Document",
    "RetrievalResult",
    "RAGResponse",
    # Retrievers
    "Retriever",
    "VectorRetriever",
    "KeywordRetriever",
    "HybridRetriever",
    "RerankingRetriever",
    # Generators
    "Generator",
    "LLMGenerator",
    # Context
    "ContextBuilder",
    # Pipeline
    "RAGPipeline",
    # Chunking
    "DocumentChunker",
    # Conversational
    "ConversationalRAG",
    # Decorators
    "with_rag",
    # Factory
    "create_rag_pipeline",
]



"""
Enterprise Ranking Module.

Provides relevance scoring, re-ranking, result ordering,
and ranking algorithms for search and retrieval.

Example:
    # Create ranker
    ranker = CrossEncoderRanker(client=openai_client)
    
    # Rank results
    results = await ranker.rank(query="best pizza", documents=[...])
    
    # Rerank existing results
    reranked = await ranker.rerank(query, search_results)
"""

from __future__ import annotations

import asyncio
import math
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
from datetime import datetime
from functools import wraps
from enum import Enum
import logging

logger = logging.getLogger(__name__)

T = TypeVar('T')


class RankingError(Exception):
    """Ranking error."""
    pass


class RankingStrategy(str, Enum):
    """Ranking strategies."""
    RELEVANCE = "relevance"       # By relevance score
    RECENCY = "recency"           # By timestamp
    POPULARITY = "popularity"     # By engagement metrics
    DIVERSITY = "diversity"       # Maximize diversity
    PERSONALIZED = "personalized" # User-specific
    HYBRID = "hybrid"             # Combination


class AggregationMethod(str, Enum):
    """Score aggregation methods."""
    SUM = "sum"
    MEAN = "mean"
    MAX = "max"
    MIN = "min"
    WEIGHTED = "weighted"
    RECIPROCAL_RANK = "rrf"


@dataclass
class RankedItem(Generic[T]):
    """A ranked item with score."""
    item: T
    score: float
    rank: int = 0
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def __lt__(self, other: 'RankedItem') -> bool:
        return self.score < other.score
    
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, RankedItem):
            return self.score == other.score
        return False


@dataclass
class RankingResult(Generic[T]):
    """Result of ranking operation."""
    items: List[RankedItem[T]]
    query: Optional[str] = None
    strategy: RankingStrategy = RankingStrategy.RELEVANCE
    duration_ms: float = 0.0
    
    @property
    def ranked_items(self) -> List[T]:
        """Get items in ranked order."""
        return [item.item for item in self.items]
    
    @property
    def scores(self) -> List[float]:
        """Get scores."""
        return [item.score for item in self.items]
    
    def top_k(self, k: int) -> List[T]:
        """Get top k items."""
        return [item.item for item in self.items[:k]]
    
    def __len__(self) -> int:
        return len(self.items)
    
    def __iter__(self):
        return iter(self.items)


@dataclass
class ScoringWeights:
    """Weights for hybrid scoring."""
    relevance: float = 1.0
    recency: float = 0.0
    popularity: float = 0.0
    personalization: float = 0.0
    
    def normalize(self) -> 'ScoringWeights':
        """Normalize weights to sum to 1."""
        total = self.relevance + self.recency + self.popularity + self.personalization
        if total == 0:
            return self
        return ScoringWeights(
            relevance=self.relevance / total,
            recency=self.recency / total,
            popularity=self.popularity / total,
            personalization=self.personalization / total,
        )


class Ranker(ABC, Generic[T]):
    """Abstract ranker interface."""
    
    @abstractmethod
    async def rank(
        self,
        items: List[T],
        query: Optional[str] = None,
        **kwargs: Any,
    ) -> RankingResult[T]:
        """Rank items."""
        pass


class ScoreBasedRanker(Ranker[T]):
    """Ranker that sorts by existing scores."""
    
    def __init__(
        self,
        score_field: str = "score",
        descending: bool = True,
    ):
        self._score_field = score_field
        self._descending = descending
    
    async def rank(
        self,
        items: List[T],
        query: Optional[str] = None,
        **kwargs: Any,
    ) -> RankingResult[T]:
        """Rank by existing scores."""
        import time
        start = time.time()
        
        ranked = []
        for item in items:
            if isinstance(item, dict):
                score = item.get(self._score_field, 0.0)
            elif hasattr(item, self._score_field):
                score = getattr(item, self._score_field, 0.0)
            else:
                score = 0.0
            
            ranked.append(RankedItem(item=item, score=score))
        
        # Sort
        ranked.sort(key=lambda x: x.score, reverse=self._descending)
        
        # Assign ranks
        for i, item in enumerate(ranked):
            item.rank = i + 1
        
        return RankingResult(
            items=ranked,
            query=query,
            strategy=RankingStrategy.RELEVANCE,
            duration_ms=(time.time() - start) * 1000,
        )


class LLMRanker(Ranker[T]):
    """LLM-based ranker for semantic relevance."""
    
    def __init__(
        self,
        client: Any,
        model: Optional[str] = None,
        content_field: str = "content",
    ):
        self._client = client
        self._model = model
        self._content_field = content_field
    
    def _get_content(self, item: T) -> str:
        """Extract content from item."""
        if isinstance(item, str):
            return item
        if isinstance(item, dict):
            return item.get(self._content_field, str(item))
        if hasattr(item, self._content_field):
            return getattr(item, self._content_field)
        return str(item)
    
    async def rank(
        self,
        items: List[T],
        query: Optional[str] = None,
        **kwargs: Any,
    ) -> RankingResult[T]:
        """Rank using LLM."""
        import time
        import json
        start = time.time()
        
        if not query:
            raise RankingError("Query required for LLM ranking")
        
        # Build prompt
        docs = []
        for i, item in enumerate(items):
            content = self._get_content(item)
            docs.append(f"[{i}] {content[:500]}")
        
        docs_str = "\n\n".join(docs)
        
        prompt = f"""Given the query and documents, rank the documents by relevance.
Return a JSON array of document indices in order of relevance (most relevant first).

Query: {query}

Documents:
{docs_str}

Return ONLY a JSON array like [2, 0, 1, 3] (no explanation):"""
        
        response = await self._client.chat.completions.create(
            model=self._model,
            messages=[{"role": "user", "content": prompt}],
            temperature=0,
        )
        
        try:
            content = response.choices[0].message.content
            import re
            match = re.search(r'\[[\d,\s]+\]', content)
            if match:
                order = json.loads(match.group())
                
                # Build ranked results
                ranked = []
                for rank, idx in enumerate(order):
                    if 0 <= idx < len(items):
                        score = 1.0 - (rank / len(order))  # Higher rank = higher score
                        ranked.append(RankedItem(
                            item=items[idx],
                            score=score,
                            rank=rank + 1,
                        ))
                
                # Add any items not in ranking
                ranked_indices = set(order)
                for i, item in enumerate(items):
                    if i not in ranked_indices:
                        ranked.append(RankedItem(
                            item=item,
                            score=0.0,
                            rank=len(ranked) + 1,
                        ))
                
                return RankingResult(
                    items=ranked,
                    query=query,
                    strategy=RankingStrategy.RELEVANCE,
                    duration_ms=(time.time() - start) * 1000,
                )
        except Exception as e:
            logger.error(f"Failed to parse LLM ranking: {e}")
        
        # Fallback to original order
        return await ScoreBasedRanker().rank(items, query)


class ReciprocalRankFusion(Ranker[T]):
    """Reciprocal Rank Fusion for combining multiple rankings."""
    
    def __init__(
        self,
        rankers: List[Ranker[T]],
        k: int = 60,
    ):
        self._rankers = rankers
        self._k = k
    
    async def rank(
        self,
        items: List[T],
        query: Optional[str] = None,
        **kwargs: Any,
    ) -> RankingResult[T]:
        """Combine rankings using RRF."""
        import time
        start = time.time()
        
        # Get rankings from all rankers
        rankings = await asyncio.gather(*[
            ranker.rank(items, query, **kwargs)
            for ranker in self._rankers
        ])
        
        # Calculate RRF scores
        scores: Dict[int, float] = {}  # item index -> score
        
        for ranking in rankings:
            for ranked_item in ranking.items:
                # Find original index
                for i, item in enumerate(items):
                    if item == ranked_item.item:
                        rrf_score = 1.0 / (self._k + ranked_item.rank)
                        scores[i] = scores.get(i, 0) + rrf_score
                        break
        
        # Sort by combined score
        sorted_indices = sorted(scores.keys(), key=lambda x: scores[x], reverse=True)
        
        ranked = []
        for rank, idx in enumerate(sorted_indices):
            ranked.append(RankedItem(
                item=items[idx],
                score=scores[idx],
                rank=rank + 1,
            ))
        
        return RankingResult(
            items=ranked,
            query=query,
            strategy=RankingStrategy.HYBRID,
            duration_ms=(time.time() - start) * 1000,
        )


class DiversityRanker(Ranker[T]):
    """Ranker that maximizes result diversity."""
    
    def __init__(
        self,
        similarity_fn: Callable[[T, T], float],
        diversity_weight: float = 0.3,
        base_ranker: Optional[Ranker[T]] = None,
    ):
        self._similarity_fn = similarity_fn
        self._diversity_weight = diversity_weight
        self._base_ranker = base_ranker or ScoreBasedRanker()
    
    async def rank(
        self,
        items: List[T],
        query: Optional[str] = None,
        **kwargs: Any,
    ) -> RankingResult[T]:
        """Rank with diversity consideration (MMR-like)."""
        import time
        start = time.time()
        
        # Get base ranking
        base_result = await self._base_ranker.rank(items, query, **kwargs)
        
        if len(items) <= 1:
            return base_result
        
        # MMR-style reranking
        selected = []
        remaining = list(range(len(base_result.items)))
        
        # Select first item (highest relevance)
        first_idx = remaining.pop(0)
        selected.append(first_idx)
        
        # Iteratively select remaining items
        while remaining:
            best_idx = None
            best_score = float('-inf')
            
            for idx in remaining:
                item = base_result.items[idx]
                relevance = item.score
                
                # Calculate max similarity to selected items
                max_similarity = 0.0
                for sel_idx in selected:
                    sel_item = base_result.items[sel_idx].item
                    
                    if asyncio.iscoroutinefunction(self._similarity_fn):
                        sim = await self._similarity_fn(item.item, sel_item)
                    else:
                        sim = self._similarity_fn(item.item, sel_item)
                    
                    max_similarity = max(max_similarity, sim)
                
                # MMR score
                mmr_score = (1 - self._diversity_weight) * relevance - self._diversity_weight * max_similarity
                
                if mmr_score > best_score:
                    best_score = mmr_score
                    best_idx = idx
            
            if best_idx is not None:
                remaining.remove(best_idx)
                selected.append(best_idx)
        
        # Build result
        ranked = []
        for rank, idx in enumerate(selected):
            orig_item = base_result.items[idx]
            ranked.append(RankedItem(
                item=orig_item.item,
                score=orig_item.score,
                rank=rank + 1,
            ))
        
        return RankingResult(
            items=ranked,
            query=query,
            strategy=RankingStrategy.DIVERSITY,
            duration_ms=(time.time() - start) * 1000,
        )


class RecencyRanker(Ranker[T]):
    """Ranker that considers recency."""
    
    def __init__(
        self,
        timestamp_field: str = "timestamp",
        decay_rate: float = 0.1,
        base_ranker: Optional[Ranker[T]] = None,
    ):
        self._timestamp_field = timestamp_field
        self._decay_rate = decay_rate
        self._base_ranker = base_ranker or ScoreBasedRanker()
    
    def _get_timestamp(self, item: T) -> datetime:
        """Extract timestamp from item."""
        if isinstance(item, dict):
            ts = item.get(self._timestamp_field)
        elif hasattr(item, self._timestamp_field):
            ts = getattr(item, self._timestamp_field)
        else:
            ts = None
        
        if isinstance(ts, datetime):
            return ts
        if isinstance(ts, str):
            return datetime.fromisoformat(ts)
        
        return datetime.now()
    
    async def rank(
        self,
        items: List[T],
        query: Optional[str] = None,
        **kwargs: Any,
    ) -> RankingResult[T]:
        """Rank with recency decay."""
        import time
        start = time.time()
        
        # Get base ranking
        base_result = await self._base_ranker.rank(items, query, **kwargs)
        
        now = datetime.now()
        ranked = []
        
        for ranked_item in base_result.items:
            ts = self._get_timestamp(ranked_item.item)
            age_hours = (now - ts).total_seconds() / 3600
            
            # Exponential decay
            recency_score = math.exp(-self._decay_rate * age_hours)
            
            # Combine scores
            combined_score = ranked_item.score * recency_score
            
            ranked.append(RankedItem(
                item=ranked_item.item,
                score=combined_score,
                metadata={"recency_score": recency_score},
            ))
        
        # Re-sort
        ranked.sort(key=lambda x: x.score, reverse=True)
        
        for i, item in enumerate(ranked):
            item.rank = i + 1
        
        return RankingResult(
            items=ranked,
            query=query,
            strategy=RankingStrategy.RECENCY,
            duration_ms=(time.time() - start) * 1000,
        )


class WeightedRanker(Ranker[T]):
    """Ranker with weighted scoring."""
    
    def __init__(
        self,
        scorers: Dict[str, Callable[[T, Optional[str]], float]],
        weights: Dict[str, float],
    ):
        self._scorers = scorers
        self._weights = weights
    
    async def rank(
        self,
        items: List[T],
        query: Optional[str] = None,
        **kwargs: Any,
    ) -> RankingResult[T]:
        """Rank using weighted scores."""
        import time
        start = time.time()
        
        ranked = []
        
        for item in items:
            scores = {}
            total_score = 0.0
            total_weight = 0.0
            
            for name, scorer in self._scorers.items():
                weight = self._weights.get(name, 1.0)
                
                if asyncio.iscoroutinefunction(scorer):
                    score = await scorer(item, query)
                else:
                    score = scorer(item, query)
                
                scores[name] = score
                total_score += score * weight
                total_weight += weight
            
            if total_weight > 0:
                final_score = total_score / total_weight
            else:
                final_score = 0.0
            
            ranked.append(RankedItem(
                item=item,
                score=final_score,
                metadata={"component_scores": scores},
            ))
        
        # Sort
        ranked.sort(key=lambda x: x.score, reverse=True)
        
        for i, item in enumerate(ranked):
            item.rank = i + 1
        
        return RankingResult(
            items=ranked,
            query=query,
            strategy=RankingStrategy.HYBRID,
            duration_ms=(time.time() - start) * 1000,
        )


class ReRanker:
    """
    Re-ranker for post-retrieval ranking.
    """
    
    def __init__(
        self,
        ranker: Ranker,
        top_k: Optional[int] = None,
    ):
        self._ranker = ranker
        self._top_k = top_k
    
    async def rerank(
        self,
        query: str,
        items: List[Any],
        k: Optional[int] = None,
    ) -> List[Any]:
        """Re-rank items and return top k."""
        result = await self._ranker.rank(items, query)
        
        k = k or self._top_k or len(items)
        return result.top_k(k)


# Decorator
def ranked(
    ranker: Ranker,
    k: Optional[int] = None,
) -> Callable:
    """
    Decorator to rank function results.
    
    Example:
        @ranked(ranker, k=10)
        async def search(query: str) -> List[Document]:
            ...
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(query: str, *args: Any, **kwargs: Any) -> RankingResult:
            items = await func(query, *args, **kwargs)
            result = await ranker.rank(items, query)
            
            if k:
                result.items = result.items[:k]
            
            return result
        
        return wrapper
    
    return decorator


# Factory functions
def create_ranker(
    strategy: str = "score",
    llm_client: Optional[Any] = None,
    model: Optional[str] = None,
    **kwargs: Any,
) -> Ranker:
    """
    Factory function to create a ranker.
    
    Args:
        strategy: Ranking strategy
        llm_client: LLM client for semantic ranking
        model: Model name
    """
    if strategy == "score":
        return ScoreBasedRanker(**kwargs)
    elif strategy == "llm":
        if not llm_client:
            raise ValueError("LLM client required for LLM ranker")
        return LLMRanker(llm_client, model, **kwargs)
    elif strategy == "recency":
        return RecencyRanker(**kwargs)
    else:
        return ScoreBasedRanker(**kwargs)


__all__ = [
    # Exceptions
    "RankingError",
    # Enums
    "RankingStrategy",
    "AggregationMethod",
    # Data classes
    "RankedItem",
    "RankingResult",
    "ScoringWeights",
    # Rankers
    "Ranker",
    "ScoreBasedRanker",
    "LLMRanker",
    "ReciprocalRankFusion",
    "DiversityRanker",
    "RecencyRanker",
    "WeightedRanker",
    # Re-ranker
    "ReRanker",
    # Decorators
    "ranked",
    # Factory
    "create_ranker",
]



"""
Enterprise Worker Pool Module.

Provides worker pool management, task distribution,
work stealing, priority queues, and parallel execution.

Example:
    # Create worker pool
    pool = create_worker_pool(num_workers=4)
    
    # Submit tasks
    result = await pool.submit(process_data, data)
    
    # Map over items
    results = await pool.map(process_item, items)
    
    # Use decorator
    @background_task()
    async def long_running_task():
        ...
"""

from __future__ import annotations

import asyncio
import logging
import time
import uuid
from abc import ABC, abstractmethod
from collections import deque
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from heapq import heappush, heappop
from typing import (
    Any,
    Callable,
    Coroutine,
    Deque,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

T = TypeVar('T')
R = TypeVar('R')


logger = logging.getLogger(__name__)


class WorkerError(Exception):
    """Base worker error."""
    pass


class TaskCancelledError(WorkerError):
    """Task was cancelled."""
    pass


class PoolShutdownError(WorkerError):
    """Pool is shutting down."""
    pass


class TaskPriority(int, Enum):
    """Task priority levels."""
    LOW = 3
    NORMAL = 2
    HIGH = 1
    CRITICAL = 0


class TaskState(str, Enum):
    """Task state."""
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"


class PoolState(str, Enum):
    """Pool state."""
    CREATED = "created"
    RUNNING = "running"
    DRAINING = "draining"
    SHUTDOWN = "shutdown"


@dataclass
class TaskConfig:
    """Task configuration."""
    priority: TaskPriority = TaskPriority.NORMAL
    timeout: Optional[float] = None
    retries: int = 0
    retry_delay: float = 1.0


@dataclass
class TaskResult(Generic[R]):
    """Task result."""
    task_id: str
    state: TaskState
    result: Optional[R] = None
    error: Optional[Exception] = None
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    duration: float = 0.0
    retries: int = 0


@dataclass
class WorkerStats:
    """Worker statistics."""
    worker_id: str
    tasks_completed: int = 0
    tasks_failed: int = 0
    total_runtime: float = 0.0
    idle_time: float = 0.0
    current_task: Optional[str] = None


@dataclass
class PoolStats:
    """Pool statistics."""
    total_workers: int = 0
    active_workers: int = 0
    idle_workers: int = 0
    pending_tasks: int = 0
    completed_tasks: int = 0
    failed_tasks: int = 0
    cancelled_tasks: int = 0


@dataclass(order=True)
class PriorityTask:
    """Priority queue task."""
    priority: int
    submitted_at: float = field(compare=False)
    task_id: str = field(compare=False)
    func: Callable = field(compare=False)
    args: Tuple = field(compare=False)
    kwargs: Dict = field(compare=False)
    config: TaskConfig = field(compare=False)


class TaskQueue(ABC):
    """Abstract task queue."""
    
    @abstractmethod
    async def put(self, task: PriorityTask) -> None:
        """Add task to queue."""
        pass
    
    @abstractmethod
    async def get(self) -> PriorityTask:
        """Get next task from queue."""
        pass
    
    @abstractmethod
    def size(self) -> int:
        """Get queue size."""
        pass


class PriorityTaskQueue(TaskQueue):
    """Priority-based task queue."""
    
    def __init__(self):
        self._heap: List[PriorityTask] = []
        self._lock = asyncio.Lock()
        self._not_empty = asyncio.Condition(self._lock)
    
    async def put(self, task: PriorityTask) -> None:
        async with self._not_empty:
            heappush(self._heap, task)
            self._not_empty.notify()
    
    async def get(self) -> PriorityTask:
        async with self._not_empty:
            while not self._heap:
                await self._not_empty.wait()
            return heappop(self._heap)
    
    def size(self) -> int:
        return len(self._heap)


class FIFOTaskQueue(TaskQueue):
    """FIFO task queue."""
    
    def __init__(self):
        self._queue: Deque[PriorityTask] = deque()
        self._lock = asyncio.Lock()
        self._not_empty = asyncio.Condition(self._lock)
    
    async def put(self, task: PriorityTask) -> None:
        async with self._not_empty:
            self._queue.append(task)
            self._not_empty.notify()
    
    async def get(self) -> PriorityTask:
        async with self._not_empty:
            while not self._queue:
                await self._not_empty.wait()
            return self._queue.popleft()
    
    def size(self) -> int:
        return len(self._queue)


class Worker:
    """
    Worker that processes tasks from queue.
    """
    
    def __init__(
        self,
        worker_id: str,
        queue: TaskQueue,
        results: Dict[str, TaskResult],
    ):
        self._id = worker_id
        self._queue = queue
        self._results = results
        self._running = False
        self._task: Optional[asyncio.Task] = None
        self._current_task_id: Optional[str] = None
        self._stats = WorkerStats(worker_id=worker_id)
        self._idle_start = time.time()
    
    @property
    def id(self) -> str:
        return self._id
    
    @property
    def stats(self) -> WorkerStats:
        return self._stats
    
    @property
    def is_busy(self) -> bool:
        return self._current_task_id is not None
    
    async def start(self) -> None:
        """Start the worker."""
        self._running = True
        self._task = asyncio.create_task(self._run())
    
    async def stop(self) -> None:
        """Stop the worker."""
        self._running = False
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
    
    async def _run(self) -> None:
        """Worker main loop."""
        while self._running:
            try:
                # Update idle time
                idle_duration = time.time() - self._idle_start
                self._stats.idle_time += idle_duration
                
                # Get next task
                priority_task = await self._queue.get()
                self._current_task_id = priority_task.task_id
                self._stats.current_task = priority_task.task_id
                
                # Execute task
                await self._execute_task(priority_task)
                
                self._current_task_id = None
                self._stats.current_task = None
                self._idle_start = time.time()
                
            except asyncio.CancelledError:
                break
            except Exception as e:
                logger.error(f"Worker {self._id} error: {e}")
    
    async def _execute_task(self, task: PriorityTask) -> None:
        """Execute a single task."""
        task_id = task.task_id
        start_time = time.time()
        retries = 0
        
        result = TaskResult(
            task_id=task_id,
            state=TaskState.RUNNING,
            started_at=datetime.utcnow(),
        )
        self._results[task_id] = result
        
        while retries <= task.config.retries:
            try:
                # Execute with timeout
                if task.config.timeout:
                    coro = task.func(*task.args, **task.kwargs)
                    if asyncio.iscoroutine(coro):
                        output = await asyncio.wait_for(
                            coro,
                            timeout=task.config.timeout,
                        )
                    else:
                        output = coro
                else:
                    output = task.func(*task.args, **task.kwargs)
                    if asyncio.iscoroutine(output):
                        output = await output
                
                # Success
                result.state = TaskState.COMPLETED
                result.result = output
                result.completed_at = datetime.utcnow()
                result.duration = time.time() - start_time
                result.retries = retries
                
                self._stats.tasks_completed += 1
                self._stats.total_runtime += result.duration
                
                return
                
            except asyncio.TimeoutError:
                retries += 1
                if retries <= task.config.retries:
                    await asyncio.sleep(task.config.retry_delay)
                else:
                    result.state = TaskState.FAILED
                    result.error = asyncio.TimeoutError("Task timed out")
                    
            except Exception as e:
                retries += 1
                if retries <= task.config.retries:
                    await asyncio.sleep(task.config.retry_delay)
                else:
                    result.state = TaskState.FAILED
                    result.error = e
        
        result.completed_at = datetime.utcnow()
        result.duration = time.time() - start_time
        result.retries = retries
        
        self._stats.tasks_failed += 1


class WorkerPool(ABC):
    """Abstract worker pool."""
    
    @abstractmethod
    async def submit(
        self,
        func: Callable[..., R],
        *args,
        priority: TaskPriority = TaskPriority.NORMAL,
        **kwargs,
    ) -> TaskResult[R]:
        """Submit a task for execution."""
        pass
    
    @abstractmethod
    async def map(
        self,
        func: Callable[[T], R],
        items: List[T],
        priority: TaskPriority = TaskPriority.NORMAL,
    ) -> List[TaskResult[R]]:
        """Map function over items."""
        pass
    
    @abstractmethod
    async def shutdown(self, wait: bool = True) -> None:
        """Shutdown the pool."""
        pass
    
    @abstractmethod
    async def stats(self) -> PoolStats:
        """Get pool statistics."""
        pass


class AsyncWorkerPool(WorkerPool):
    """
    Async worker pool implementation.
    """
    
    def __init__(
        self,
        num_workers: int = 4,
        queue: Optional[TaskQueue] = None,
    ):
        self._num_workers = num_workers
        self._queue = queue or PriorityTaskQueue()
        self._workers: List[Worker] = []
        self._results: Dict[str, TaskResult] = {}
        self._state = PoolState.CREATED
        self._stats = PoolStats(total_workers=num_workers)
    
    async def start(self) -> None:
        """Start the worker pool."""
        if self._state != PoolState.CREATED:
            return
        
        self._state = PoolState.RUNNING
        
        for i in range(self._num_workers):
            worker = Worker(
                worker_id=f"worker-{i}",
                queue=self._queue,
                results=self._results,
            )
            self._workers.append(worker)
            await worker.start()
    
    async def submit(
        self,
        func: Callable[..., R],
        *args,
        priority: TaskPriority = TaskPriority.NORMAL,
        timeout: Optional[float] = None,
        retries: int = 0,
        **kwargs,
    ) -> TaskResult[R]:
        if self._state not in (PoolState.CREATED, PoolState.RUNNING):
            raise PoolShutdownError("Pool is shutting down")
        
        if self._state == PoolState.CREATED:
            await self.start()
        
        task_id = str(uuid.uuid4())
        config = TaskConfig(
            priority=priority,
            timeout=timeout,
            retries=retries,
        )
        
        priority_task = PriorityTask(
            priority=priority.value,
            submitted_at=time.time(),
            task_id=task_id,
            func=func,
            args=args,
            kwargs=kwargs,
            config=config,
        )
        
        # Create pending result
        self._results[task_id] = TaskResult(
            task_id=task_id,
            state=TaskState.PENDING,
        )
        
        await self._queue.put(priority_task)
        self._stats.pending_tasks += 1
        
        # Wait for completion
        while self._results[task_id].state in (TaskState.PENDING, TaskState.RUNNING):
            await asyncio.sleep(0.01)
        
        result = self._results[task_id]
        
        if result.state == TaskState.COMPLETED:
            self._stats.completed_tasks += 1
        elif result.state == TaskState.FAILED:
            self._stats.failed_tasks += 1
        
        self._stats.pending_tasks = max(0, self._stats.pending_tasks - 1)
        
        return result
    
    async def submit_nowait(
        self,
        func: Callable[..., R],
        *args,
        priority: TaskPriority = TaskPriority.NORMAL,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> str:
        """Submit task without waiting for result."""
        if self._state == PoolState.CREATED:
            await self.start()
        
        task_id = str(uuid.uuid4())
        config = TaskConfig(priority=priority, timeout=timeout)
        
        priority_task = PriorityTask(
            priority=priority.value,
            submitted_at=time.time(),
            task_id=task_id,
            func=func,
            args=args,
            kwargs=kwargs,
            config=config,
        )
        
        self._results[task_id] = TaskResult(
            task_id=task_id,
            state=TaskState.PENDING,
        )
        
        await self._queue.put(priority_task)
        
        return task_id
    
    async def get_result(self, task_id: str, timeout: Optional[float] = None) -> TaskResult:
        """Get result for task ID."""
        start = time.time()
        
        while True:
            result = self._results.get(task_id)
            
            if result and result.state not in (TaskState.PENDING, TaskState.RUNNING):
                return result
            
            if timeout and (time.time() - start) > timeout:
                raise asyncio.TimeoutError()
            
            await asyncio.sleep(0.01)
    
    async def map(
        self,
        func: Callable[[T], R],
        items: List[T],
        priority: TaskPriority = TaskPriority.NORMAL,
    ) -> List[TaskResult[R]]:
        # Submit all tasks
        task_ids = []
        for item in items:
            task_id = await self.submit_nowait(func, item, priority=priority)
            task_ids.append(task_id)
        
        # Wait for all results
        results = []
        for task_id in task_ids:
            result = await self.get_result(task_id)
            results.append(result)
        
        return results
    
    async def cancel(self, task_id: str) -> bool:
        """Cancel a pending task."""
        result = self._results.get(task_id)
        
        if result and result.state == TaskState.PENDING:
            result.state = TaskState.CANCELLED
            self._stats.cancelled_tasks += 1
            return True
        
        return False
    
    async def shutdown(self, wait: bool = True) -> None:
        """Shutdown the pool."""
        self._state = PoolState.DRAINING
        
        if wait:
            # Wait for pending tasks
            while self._queue.size() > 0:
                await asyncio.sleep(0.1)
        
        self._state = PoolState.SHUTDOWN
        
        # Stop all workers
        for worker in self._workers:
            await worker.stop()
    
    async def stats(self) -> PoolStats:
        active = sum(1 for w in self._workers if w.is_busy)
        
        return PoolStats(
            total_workers=len(self._workers),
            active_workers=active,
            idle_workers=len(self._workers) - active,
            pending_tasks=self._queue.size(),
            completed_tasks=self._stats.completed_tasks,
            failed_tasks=self._stats.failed_tasks,
            cancelled_tasks=self._stats.cancelled_tasks,
        )


class WorkerPoolRegistry:
    """Registry for worker pools."""
    
    def __init__(self):
        self._pools: Dict[str, WorkerPool] = {}
        self._default: Optional[str] = None
    
    def register(
        self,
        name: str,
        pool: WorkerPool,
        default: bool = False,
    ) -> None:
        self._pools[name] = pool
        if default or self._default is None:
            self._default = name
    
    def get(self, name: Optional[str] = None) -> WorkerPool:
        name = name or self._default
        if not name or name not in self._pools:
            raise WorkerError(f"Pool not found: {name}")
        return self._pools[name]


# Global registry
_global_registry = WorkerPoolRegistry()


# Decorators
def background_task(
    priority: TaskPriority = TaskPriority.NORMAL,
    timeout: Optional[float] = None,
    pool_name: Optional[str] = None,
) -> Callable:
    """
    Decorator to run function as background task.
    
    Example:
        @background_task(priority=TaskPriority.HIGH)
        async def process_data(data):
            ...
    """
    def decorator(func: Callable) -> Callable:
        async def wrapper(*args, **kwargs):
            pool = get_worker_pool(pool_name)
            return await pool.submit(
                func, *args,
                priority=priority,
                timeout=timeout,
                **kwargs
            )
        
        wrapper._original = func
        return wrapper
    
    return decorator


def parallel(num_workers: int = 4) -> Callable:
    """
    Decorator to run function with parallel workers.
    
    Example:
        @parallel(num_workers=8)
        async def process_batch(items):
            ...
    """
    def decorator(func: Callable) -> Callable:
        pool = AsyncWorkerPool(num_workers=num_workers)
        
        async def wrapper(items: List, *args, **kwargs):
            if pool._state == PoolState.CREATED:
                await pool.start()
            
            async def process(item):
                return await func(item, *args, **kwargs)
            
            results = await pool.map(process, items)
            return [r.result for r in results if r.state == TaskState.COMPLETED]
        
        wrapper._pool = pool
        return wrapper
    
    return decorator


# Factory functions
def create_worker_pool(
    num_workers: int = 4,
    priority_queue: bool = True,
) -> AsyncWorkerPool:
    """Create a worker pool."""
    queue = PriorityTaskQueue() if priority_queue else FIFOTaskQueue()
    return AsyncWorkerPool(num_workers=num_workers, queue=queue)


def create_task_config(
    priority: TaskPriority = TaskPriority.NORMAL,
    timeout: Optional[float] = None,
    retries: int = 0,
    retry_delay: float = 1.0,
) -> TaskConfig:
    """Create a task configuration."""
    return TaskConfig(
        priority=priority,
        timeout=timeout,
        retries=retries,
        retry_delay=retry_delay,
    )


def register_worker_pool(
    name: str,
    pool: WorkerPool,
    default: bool = False,
) -> None:
    """Register pool in global registry."""
    _global_registry.register(name, pool, default)


def get_worker_pool(name: Optional[str] = None) -> WorkerPool:
    """Get pool from global registry."""
    try:
        return _global_registry.get(name)
    except WorkerError:
        pool = create_worker_pool()
        register_worker_pool("default", pool, default=True)
        return pool


__all__ = [
    # Exceptions
    "WorkerError",
    "TaskCancelledError",
    "PoolShutdownError",
    # Enums
    "TaskPriority",
    "TaskState",
    "PoolState",
    # Data classes
    "TaskConfig",
    "TaskResult",
    "WorkerStats",
    "PoolStats",
    "PriorityTask",
    # Queue
    "TaskQueue",
    "PriorityTaskQueue",
    "FIFOTaskQueue",
    # Worker
    "Worker",
    # Pool
    "WorkerPool",
    "AsyncWorkerPool",
    # Registry
    "WorkerPoolRegistry",
    # Decorators
    "background_task",
    "parallel",
    # Factory functions
    "create_worker_pool",
    "create_task_config",
    "register_worker_pool",
    "get_worker_pool",
]



"""
Enterprise PDF Generator Module.

Provides PDF generation from HTML templates,
document merging, watermarking, and styling.

Example:
    # Create PDF generator
    pdf = create_pdf_generator()
    
    # Generate from HTML
    content = await pdf.from_html(
        '<h1>Hello World</h1>',
        options={'page-size': 'A4'},
    )
    
    # Generate from template
    content = await pdf.from_template(
        template='invoice',
        context={'items': [...], 'total': 100},
    )
    
    # Save to file
    await pdf.save(content, 'output.pdf')
"""

from __future__ import annotations

import asyncio
import base64
import functools
import logging
import os
import tempfile
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class PDFError(Exception):
    """PDF generation error."""
    pass


class TemplateError(PDFError):
    """Template rendering error."""
    pass


class RenderError(PDFError):
    """PDF rendering error."""
    pass


class PageSize(str, Enum):
    """Page sizes."""
    A3 = "A3"
    A4 = "A4"
    A5 = "A5"
    LETTER = "Letter"
    LEGAL = "Legal"
    TABLOID = "Tabloid"


class Orientation(str, Enum):
    """Page orientation."""
    PORTRAIT = "portrait"
    LANDSCAPE = "landscape"


@dataclass
class Margins:
    """Page margins in mm."""
    top: int = 10
    right: int = 10
    bottom: int = 10
    left: int = 10


@dataclass
class PageOptions:
    """PDF page options."""
    size: PageSize = PageSize.A4
    orientation: Orientation = Orientation.PORTRAIT
    margins: Margins = field(default_factory=Margins)
    header: str = ""
    footer: str = ""
    header_height: int = 0
    footer_height: int = 0


@dataclass
class PDFOptions:
    """PDF generation options."""
    page: PageOptions = field(default_factory=PageOptions)
    title: str = ""
    author: str = ""
    subject: str = ""
    keywords: List[str] = field(default_factory=list)
    compress: bool = True
    javascript: bool = False
    print_background: bool = True
    prefer_css_page_size: bool = False


@dataclass
class WatermarkOptions:
    """Watermark options."""
    text: str = ""
    image_path: str = ""
    opacity: float = 0.5
    rotation: int = 45
    font_size: int = 48
    color: str = "#888888"
    position: str = "center"  # center, top, bottom


@dataclass
class PDFDocument:
    """PDF document."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    content: bytes = b""
    pages: int = 0
    size: int = 0
    title: str = ""
    created_at: datetime = field(default_factory=datetime.utcnow)
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def to_base64(self) -> str:
        """Convert to base64."""
        return base64.b64encode(self.content).decode()
    
    def to_data_uri(self) -> str:
        """Convert to data URI."""
        return f"data:application/pdf;base64,{self.to_base64()}"


# Template engine
class TemplateEngine(ABC):
    """Template engine."""
    
    @abstractmethod
    def render(
        self,
        template: str,
        context: Dict[str, Any],
    ) -> str:
        """Render template."""
        pass


class SimpleTemplateEngine(TemplateEngine):
    """Simple template engine."""
    
    def __init__(
        self,
        templates: Optional[Dict[str, str]] = None,
    ):
        self._templates = templates or {}
    
    def add_template(self, name: str, content: str) -> None:
        """Add template."""
        self._templates[name] = content
    
    def render(
        self,
        template: str,
        context: Dict[str, Any],
    ) -> str:
        if template in self._templates:
            template = self._templates[template]
        
        result = template
        for key, value in context.items():
            result = result.replace(f"{{{{{key}}}}}", str(value))
        
        return result


# PDF renderers
class PDFRenderer(ABC):
    """Abstract PDF renderer."""
    
    @abstractmethod
    async def render_html(
        self,
        html: str,
        options: PDFOptions,
    ) -> bytes:
        """Render HTML to PDF."""
        pass


class MockRenderer(PDFRenderer):
    """Mock renderer for testing."""
    
    async def render_html(
        self,
        html: str,
        options: PDFOptions,
    ) -> bytes:
        # Generate minimal valid PDF
        pdf_content = b"""%PDF-1.4
1 0 obj
<< /Type /Catalog /Pages 2 0 R >>
endobj
2 0 obj
<< /Type /Pages /Kids [3 0 R] /Count 1 >>
endobj
3 0 obj
<< /Type /Page /Parent 2 0 R /MediaBox [0 0 612 792] >>
endobj
xref
0 4
0000000000 65535 f 
0000000009 00000 n 
0000000058 00000 n 
0000000115 00000 n 
trailer
<< /Size 4 /Root 1 0 R >>
startxref
196
%%EOF"""
        return pdf_content


class PDFGenerator:
    """
    PDF generation service.
    """
    
    def __init__(
        self,
        renderer: Optional[PDFRenderer] = None,
        template_engine: Optional[TemplateEngine] = None,
        default_options: Optional[PDFOptions] = None,
    ):
        self._renderer = renderer or MockRenderer()
        self._template_engine = template_engine or SimpleTemplateEngine()
        self._default_options = default_options or PDFOptions()
        
        self._styles: Dict[str, str] = {}
    
    def add_template(self, name: str, content: str) -> None:
        """Add HTML template."""
        if isinstance(self._template_engine, SimpleTemplateEngine):
            self._template_engine.add_template(name, content)
    
    def add_style(self, name: str, css: str) -> None:
        """Add CSS style."""
        self._styles[name] = css
    
    async def from_html(
        self,
        html: str,
        options: Optional[PDFOptions] = None,
        styles: Optional[List[str]] = None,
    ) -> PDFDocument:
        """
        Generate PDF from HTML.
        
        Args:
            html: HTML content
            options: PDF options
            styles: Style names to include
            
        Returns:
            PDF document
        """
        options = options or self._default_options
        
        # Add styles
        style_content = ""
        for style_name in (styles or []):
            if style_name in self._styles:
                style_content += f"<style>{self._styles[style_name]}</style>"
        
        # Wrap HTML if needed
        if not html.strip().lower().startswith("<!doctype"):
            html = f"""<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <title>{options.title}</title>
    {style_content}
</head>
<body>
{html}
</body>
</html>"""
        elif style_content:
            html = html.replace("</head>", f"{style_content}</head>")
        
        # Render
        content = await self._renderer.render_html(html, options)
        
        return PDFDocument(
            content=content,
            pages=1,  # Would be calculated in real implementation
            size=len(content),
            title=options.title,
        )
    
    async def from_template(
        self,
        template: str,
        context: Dict[str, Any],
        options: Optional[PDFOptions] = None,
        styles: Optional[List[str]] = None,
    ) -> PDFDocument:
        """
        Generate PDF from template.
        
        Args:
            template: Template name or content
            context: Template context
            options: PDF options
            styles: Style names to include
            
        Returns:
            PDF document
        """
        html = self._template_engine.render(template, context)
        return await self.from_html(html, options, styles)
    
    async def from_url(
        self,
        url: str,
        options: Optional[PDFOptions] = None,
    ) -> PDFDocument:
        """
        Generate PDF from URL.
        
        Args:
            url: URL to render
            options: PDF options
            
        Returns:
            PDF document
        """
        # Would fetch and render URL in real implementation
        return await self.from_html(f"<p>Content from {url}</p>", options)
    
    async def merge(
        self,
        documents: List[PDFDocument],
        options: Optional[PDFOptions] = None,
    ) -> PDFDocument:
        """
        Merge multiple PDF documents.
        
        Args:
            documents: List of documents to merge
            options: PDF options
            
        Returns:
            Merged document
        """
        if not documents:
            raise PDFError("No documents to merge")
        
        # Would use PyPDF2 or similar in real implementation
        merged_content = b"".join(doc.content for doc in documents)
        total_pages = sum(doc.pages for doc in documents)
        
        return PDFDocument(
            content=merged_content,
            pages=total_pages,
            size=len(merged_content),
        )
    
    async def add_watermark(
        self,
        document: PDFDocument,
        watermark: WatermarkOptions,
    ) -> PDFDocument:
        """
        Add watermark to PDF.
        
        Args:
            document: PDF document
            watermark: Watermark options
            
        Returns:
            Watermarked document
        """
        # Would use PyPDF2 or similar in real implementation
        return PDFDocument(
            content=document.content,
            pages=document.pages,
            size=document.size,
            title=document.title,
            metadata={**document.metadata, "watermarked": True},
        )
    
    async def add_page_numbers(
        self,
        document: PDFDocument,
        format_string: str = "Page {page} of {total}",
        position: str = "bottom-center",
    ) -> PDFDocument:
        """
        Add page numbers to PDF.
        
        Args:
            document: PDF document
            format_string: Page number format
            position: Position on page
            
        Returns:
            Document with page numbers
        """
        # Would use PyPDF2 or similar in real implementation
        return PDFDocument(
            content=document.content,
            pages=document.pages,
            size=document.size,
            title=document.title,
            metadata={**document.metadata, "page_numbers": True},
        )
    
    async def split(
        self,
        document: PDFDocument,
        page_ranges: Optional[List[tuple]] = None,
    ) -> List[PDFDocument]:
        """
        Split PDF into multiple documents.
        
        Args:
            document: PDF document
            page_ranges: List of (start, end) tuples
            
        Returns:
            List of split documents
        """
        if not page_ranges:
            # Split into individual pages
            page_ranges = [(i, i) for i in range(1, document.pages + 1)]
        
        # Would use PyPDF2 or similar in real implementation
        return [
            PDFDocument(
                content=document.content,
                pages=end - start + 1,
                size=len(document.content),
            )
            for start, end in page_ranges
        ]
    
    async def encrypt(
        self,
        document: PDFDocument,
        password: str,
        permissions: Optional[Dict[str, bool]] = None,
    ) -> PDFDocument:
        """
        Encrypt PDF with password.
        
        Args:
            document: PDF document
            password: Password
            permissions: Permission flags
            
        Returns:
            Encrypted document
        """
        # Would use PyPDF2 or similar in real implementation
        return PDFDocument(
            content=document.content,
            pages=document.pages,
            size=document.size,
            title=document.title,
            metadata={**document.metadata, "encrypted": True},
        )
    
    async def compress(
        self,
        document: PDFDocument,
    ) -> PDFDocument:
        """
        Compress PDF.
        
        Args:
            document: PDF document
            
        Returns:
            Compressed document
        """
        # Would use gs or similar in real implementation
        return document
    
    async def save(
        self,
        document: PDFDocument,
        path: Union[str, Path],
    ) -> None:
        """Save PDF to file."""
        Path(path).write_bytes(document.content)


# Decorators
def pdf_response(
    template: str,
    filename: str = "document.pdf",
) -> Callable:
    """Decorator to return PDF response."""
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            context = await func(*args, **kwargs)
            
            generator = create_pdf_generator()
            doc = await generator.from_template(template, context)
            
            return {
                "content": doc.content,
                "filename": filename,
                "content_type": "application/pdf",
            }
        return wrapper
    return decorator


# Factory functions
def create_pdf_generator(
    renderer: Optional[PDFRenderer] = None,
    template_engine: Optional[TemplateEngine] = None,
    default_options: Optional[PDFOptions] = None,
) -> PDFGenerator:
    """Create PDF generator."""
    return PDFGenerator(
        renderer=renderer,
        template_engine=template_engine,
        default_options=default_options,
    )


def create_pdf_options(
    size: PageSize = PageSize.A4,
    orientation: Orientation = Orientation.PORTRAIT,
    margins: Optional[Margins] = None,
    title: str = "",
    **kwargs,
) -> PDFOptions:
    """Create PDF options."""
    page = PageOptions(
        size=size,
        orientation=orientation,
        margins=margins or Margins(),
    )
    return PDFOptions(page=page, title=title, **kwargs)


def create_margins(
    top: int = 10,
    right: int = 10,
    bottom: int = 10,
    left: int = 10,
) -> Margins:
    """Create page margins."""
    return Margins(top=top, right=right, bottom=bottom, left=left)


def create_watermark(
    text: str = "",
    opacity: float = 0.5,
    rotation: int = 45,
    **kwargs,
) -> WatermarkOptions:
    """Create watermark options."""
    return WatermarkOptions(
        text=text,
        opacity=opacity,
        rotation=rotation,
        **kwargs,
    )


def create_template_engine(
    templates: Optional[Dict[str, str]] = None,
) -> SimpleTemplateEngine:
    """Create template engine."""
    return SimpleTemplateEngine(templates)


def create_mock_renderer() -> MockRenderer:
    """Create mock renderer for testing."""
    return MockRenderer()


__all__ = [
    # Exceptions
    "PDFError",
    "TemplateError",
    "RenderError",
    # Enums
    "PageSize",
    "Orientation",
    # Data classes
    "Margins",
    "PageOptions",
    "PDFOptions",
    "WatermarkOptions",
    "PDFDocument",
    # Template engine
    "TemplateEngine",
    "SimpleTemplateEngine",
    # Renderers
    "PDFRenderer",
    "MockRenderer",
    # Generator
    "PDFGenerator",
    # Decorators
    "pdf_response",
    # Factory functions
    "create_pdf_generator",
    "create_pdf_options",
    "create_margins",
    "create_watermark",
    "create_template_engine",
    "create_mock_renderer",
]
